"""
This file contains various values used to avoid magic numbers and strings in
the application.
"""
# pylint: disable=too-few-public-methods


class Version:
    """
    The SNMP Version identifier. This is used in the SNMP :term:`PDU`.
    """

    V2C = 0x01
    V1 = 0x00


class Length:
    """
    A simple "namespace" to avoid magic values for indefinite lengths.
    """

    INDEFINITE = "indefinite"

#: Maximum number of usable varbinds as defined in RFC 3416
MAX_VARBINDS = 2147483647

#: A magic value used to detect strict error-handling
ERRORS_STRICT = 'strict'

#: A magic value used to detect lenient error-handling
ERRORS_WARN = 'warn'

#: TCP timeout which is used if not manually overridden
DEFAULT_TIMEOUT = 6



# pylint: skip-file

"""
Test the external "raw" interface.

"raw" means that types are not converted to pythonic types but are returned as
PureSNMP object instances.
"""

import sys
from datetime import timedelta
from unittest import skipUnless

import pytest

from puresnmp.aio.api.raw import (
    bulkget,
    bulktable,
    bulkwalk,
    get,
    getnext,
    multiget,
    multiset,
    multiwalk,
    set,
    walk
)
from puresnmp.const import Version
from puresnmp.exc import NoSuchOID, SnmpError
from puresnmp.pdu import BulkGetRequest, GetNextRequest, GetRequest, VarBind
from puresnmp.types import Counter, Gauge, IpAddress, TimeTicks
from puresnmp.util import BulkResult
from puresnmp.x690.types import (
    Integer,
    ObjectIdentifier,
    OctetString,
    Sequence,
    to_bytes
)

from . import readbytes, readbytes_multiple
from .asyncmock import AsyncMock

try:
    from unittest.mock import patch, call
except ImportError:
    from mock import patch, call  # type: ignore

pytestmark = pytest.mark.skipif(sys.version_info < (3, 5),
                                reason="requires python3.5")


class TestGet(object):

    @pytest.mark.asyncio
    async def test_get_string(self):
        data = readbytes('get_sysdescr_01.hex')
        expected = OctetString(
            b'Linux d24cf7f36138 4.4.0-28-generic #47-Ubuntu SMP '
            b'Fri Jun 24 10:09:13 UTC 2016 x86_64')
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            result = await get('::1', 'private', '1.2.3')
        assert result == expected

    @pytest.mark.asyncio
    async def test_get_oid(self):
        data = readbytes('get_sysoid_01.hex')
        expected = ObjectIdentifier.from_string('1.3.6.1.4.1.8072.3.2.10')
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            result = await get('::1', 'private', '1.2.3')
        assert result == expected

    @pytest.mark.asyncio
    async def test_get_multiple_return_binds(self):
        """
        A "GET" response should only return one varbind.
        """
        data = readbytes('get_sysoid_01_error.hex')
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            with pytest.raises(SnmpError, match='varbind'):
                await get('::1', 'private', '1.2.3')

    @pytest.mark.asyncio
    async def test_get_non_existing_oid(self):
        """
        A "GET" response on a non-existing OID should raise an appropriate
        exception.
        """
        data = readbytes('get_non_existing.hex')
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            with pytest.raises(NoSuchOID):
                await get('::1', 'private', '1.2.3')


class TestWalk(object):

    @pytest.mark.asyncio
    async def test_walk(self):
        response_1 = readbytes('walk_response_1.hex')
        response_2 = readbytes('walk_response_2.hex')
        response_3 = readbytes('walk_response_3.hex')

        expected = [VarBind(
            ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.5.1'),
            Gauge(10000000)
        ), VarBind(
            ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.5.13'),
            Gauge(4294967295)
        )]

        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().send.side_effect = [response_1, response_2, response_3]
            mck().get_request_id.return_value = 0
            result = []
            async for x in walk('::1', 'public', '1.3.6.1.2.1.2.2.1.5'):
                result.append(x)
        assert result == expected

    @pytest.mark.asyncio
    async def test_walk_multiple_return_binds(self):
        """
        A "WALK" response should only return one varbind.
        """
        data = readbytes('get_sysoid_01_error.hex')
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            with pytest.raises(SnmpError, match='varbind'):
                async for x in walk('::1', 'private', '1.2.3'):
                    pass


class TestBulkTable(object):

    @pytest.mark.asyncio
    async def test_bulktable(self):
        responses = readbytes_multiple('bulktable_response.hex')

        expected = [{
            '0': '1',
            '1': Integer(1),
            '2': OctetString(b'lo'),
            '3': Integer(24),
            '4': Integer(65536),
            '5': Gauge(10000000),
            '6': OctetString(b''),
            '7': Integer(1),
            '8': Integer(1),
            '9': TimeTicks(0),
            '10': Counter(172),
            '11': Counter(2),
            '12': Counter(0),
            '13': Counter(0),
            '14': Counter(0),
            '15': Counter(0),
            '16': Counter(172),
            '17': Counter(2),
            '18': Counter(0),
            '19': Counter(0),
            '20': Counter(0),
            '21': Gauge(0),
            '22': ObjectIdentifier(0, 0),
        }, {
            '0': '4',
            '1': Integer(4),
            '2': OctetString(b'eth0'),
            '3': Integer(6),
            '4': Integer(1500),
            '5': Gauge(4294967295),
            '6': OctetString(b'\x02B\xac\x11\x00\x02'),
            '7': Integer(1),
            '8': Integer(1),
            '9': TimeTicks(0),
            '10': Counter(548760),
            '11': Counter(3888),
            '12': Counter(0),
            '13': Counter(0),
            '14': Counter(0),
            '15': Counter(0),
            '16': Counter(186660),
            '17': Counter(1875),
            '18': Counter(0),
            '19': Counter(0),
            '20': Counter(0),
            '21': Gauge(0),
            '22': ObjectIdentifier(0, 0),
        }]

        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().send.side_effect = responses
            mck().get_request_id.return_value = 0
            result = []
            table = await bulktable('127.0.0.1', 'private', '1.3.6.1.2.1.2.2')
            for row in table:
                result.append(row)
        assert sorted(result, key=lambda x: x['0']) == expected


class TestMultiGet(object):

    @pytest.mark.asyncio
    async def test_multiget(self):
        data = readbytes('multiget_response.hex')
        expected = [
            ObjectIdentifier.from_string('1.3.6.1.4.1.8072.3.2.10'),
            OctetString(b"Linux 7fbf2f0c363d 4.4.0-28-generic "
                        b"#47-Ubuntu SMP Fri Jun 24 10:09:13 "
                        b"UTC 2016 x86_64")
        ]
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().get_request_id.return_value = 0
            mck().send.return_value = data
            result = await multiget('::1', 'private', [
                '1.3.6.1.2.1.1.2.0',
                '1.3.6.1.2.1.1.1.0',
            ])
        assert result == expected


class TestMultiWalk(object):

    # TODO (advanced) figure out why this fails on Python 2!
    @skipUnless(
        sys.version_info >= (3, 0),
        "For some reason this fails on Python 2 and I don't know why")
    @pytest.mark.asyncio
    async def test_multi_walk(self):
        response_1 = readbytes('multiwalk_response_1.hex')
        response_2 = readbytes('multiwalk_response_2.hex')
        response_3 = readbytes('multiwalk_response_3.hex')

        expected = [VarBind(
            ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.1.1'), Integer(1)
        ), VarBind(
            ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.2.1'),
            OctetString(b'lo')
        ), VarBind(
            ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.1.78'), Integer(78)
        ), VarBind(
            ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.2.78'),
            OctetString(b'eth0')
        )]

        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().get_request_id.return_value = 0
            mck().send.side_effect = [response_1, response_2, response_3]
            result = []
            async for x in multiwalk('::1', 'public', [
                '1.3.6.1.2.1.2.2.1.1',
                '1.3.6.1.2.1.2.2.1.2'
            ]):
                result.append(x)
        # TODO (advanced): should order matter in the following result?
        assert len(result) == len(expected)

    @pytest.mark.asyncio
    async def test_eom(self):
        '''
        A test for a walk operation which runs into the endOfMibView marker
        '''

        data_generator = readbytes_multiple('x690/multiwalk_endofmibview.hex')

        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().get_request_id.return_value = 0
            mck().send.side_effect = data_generator
            result = []
            async for row in multiwalk('::1', 'public', [
                '1.3.6.1.6.3.16.1.5.2.1.6.6.95.110.111.110.101.95.1',
            ]):
                result.append(row)

        root = '1.3.6.1.6.3.16.1.5.2.1.6.6.95.110.111.110.101.95.1.'
        expected = [
            (root+'0', 1),
            (root+'1', 1),
            (root+'2', 1),
        ]

        simplified_result = [
            (str(oid), value.pythonize()) for oid, value in result
        ]

        assert simplified_result == expected


class TestMultiSet(object):

    @pytest.mark.asyncio
    async def test_multiset(self):
        """
        Test setting multiple OIDs at once.

        NOTE: The OID '1.3.6.1.2.1.1.5.0' below is manually edited for
              unit-testing. It probably has a different type in the real world!
        """
        data = readbytes('multiset_response.hex')
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().get_request_id.return_value = 0
            mck().send.return_value = data
            result = await multiset('::1', 'private', [
                ('1.3.6.1.2.1.1.4.0', OctetString(b'hello@world.com')),
                ('1.3.6.1.2.1.1.5.0', OctetString(b'hello@world.com')),
            ])
        expected = {
            '1.3.6.1.2.1.1.4.0': OctetString(b'hello@world.com'),
            '1.3.6.1.2.1.1.5.0': OctetString(b'hello@world.com'),
        }
        assert result == expected


class TestGetNext(object):

    @pytest.mark.asyncio
    async def test_get_call_args(self):
        data = readbytes('dummy.hex')  # any dump would do
        packet = Sequence(
            Integer(Version.V2C),
            OctetString('public'),
            GetNextRequest(0, ObjectIdentifier(1, 2, 3))
        )
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().get_request_id.return_value = 0
            mck().send.return_value = data
            await getnext('::1', 'public', '1.2.3')
            mck.assert_called_with(timeout=6)
            mck().send.assert_called_with(
                '::1', 161, to_bytes(packet))

    @pytest.mark.asyncio
    async def test_getnext(self):
        data = readbytes('getnext_response.hex')
        expected = VarBind('1.3.6.1.6.3.1.1.6.1.0', Integer(354522558))

        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().get_request_id.return_value = 0
            mck().send.return_value = data
            result = await getnext('::1', 'private', '1.3.6.1.5')
        assert result == expected


class TestGetBulkGet(object):

    @pytest.mark.asyncio
    async def test_get_call_args(self):
        data = readbytes('dummy.hex')  # any dump would do
        packet = Sequence(
            Integer(Version.V2C),
            OctetString('public'),
            BulkGetRequest(0, 1, 2,
                           ObjectIdentifier(1, 2, 3),
                           ObjectIdentifier(1, 2, 4))
        )
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().get_request_id.return_value = 0
            mck().send.return_value = data
            await bulkget('::1', 'public',
                          ['1.2.3'],
                          ['1.2.4'],
                          max_list_size=2)
            mck.assert_called_with(timeout=6)
            mck().send.assert_called_with('::1', 161, to_bytes(packet))

    @pytest.mark.asyncio
    async def test_bulkget(self):
        data = readbytes('bulk_get_response.hex')
        expected = BulkResult(
            {'1.3.6.1.2.1.1.1.0': OctetString(
                b'Linux 7e68e60fe303 4.4.0-28-generic '
                b'#47-Ubuntu SMP Fri Jun 24 10:09:13 UTC 2016 x86_64')},
            {'1.3.6.1.2.1.3.1.1.1.10.1.172.17.0.1': Integer(10),
             '1.3.6.1.2.1.3.1.1.2.10.1.172.17.0.1':
             OctetString(b'\x02B\xe2\xc5\x8d\t'),
             '1.3.6.1.2.1.3.1.1.3.10.1.172.17.0.1':
             IpAddress(b'\xac\x11\x00\x01'),
             '1.3.6.1.2.1.4.1.0': Integer(1),
             '1.3.6.1.2.1.4.3.0': Counter(57)})

        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().get_request_id.return_value = 0
            mck().send.return_value = data
            result = await bulkget('::1', 'public',
                                   ['1.3.6.1.2.1.1.1'],
                                   ['1.3.6.1.2.1.3.1'],
                                   max_list_size=5)
        assert result == expected

    @pytest.mark.asyncio
    async def test_eom(self):
        '''
        Test a bulg-get operation which runs into the "endOfMibView" marker.
        '''

        data = readbytes('x690/bulk_get_eom_response.hex')
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().get_request_id.return_value = 0
            mck().send.return_value = data
            result = await bulkget('::1', 'public', [], ['1.2.4'],
                                   max_list_size=10)

        expected_scalars = {}
        assert result.scalars == expected_scalars

        root = '1.3.6.1.6.3.16.1.5.2.1.6.6.95.110.111.110.101.95.1.'
        expected_listing = {
            root + '0': Integer(1),
            root + '1': Integer(1),
            root + '2': Integer(1),
        }
        assert result.listing == expected_listing


class TestGetBulkWalk(object):

    @pytest.mark.asyncio
    async def test_get_call_args(self):
        data = readbytes('dummy.hex')  # any dump would do
        packet = Sequence(
            Integer(Version.V2C),
            OctetString('public'),
            BulkGetRequest(0, 0, 2, ObjectIdentifier(1, 2, 3))
        )
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().get_request_id.return_value = 0
            mck().send.return_value = data

            # we need to wrap this in a list to consume the generator.
            async for x in bulkwalk('::1', 'public',
                                    ['1.2.3'],
                                    bulk_size=2):
                pass
            mck.assert_called_with(timeout=6)
            mck().send.assert_called_with('::1', 161, to_bytes(packet))

    @pytest.mark.asyncio
    async def test_get_call_args_issue_22(self):
        data = readbytes('dummy.hex')  # any dump would do
        packet = Sequence(
            Integer(Version.V2C),
            OctetString('public'),
            BulkGetRequest(0, 0, 2, ObjectIdentifier(1, 2, 3))
        )
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().send.return_value = data
            mck().get_request_id.return_value = 0

            with pytest.raises(TypeError, match=r'OIDS.*list'):
                # we need to wrap this in a list to consume the generator.
                async for x in bulkwalk('::1', 'public', '1.2.3', bulk_size=2):
                    pass

    @pytest.mark.asyncio
    async def test_bulkwalk(self):
        req1 = readbytes('bulkwalk_request_1.hex')
        req2 = readbytes('bulkwalk_request_2.hex')
        req3 = readbytes('bulkwalk_request_3.hex')

        responses = [
            readbytes('bulkwalk_response_1.hex'),
            readbytes('bulkwalk_response_2.hex'),
            readbytes('bulkwalk_response_3.hex'),
        ]
        with patch('puresnmp.aio.api.raw.Transport') as mck:
            mck().send = AsyncMock()
            mck().send.side_effect = responses

            request_ids = [1001613222, 1001613223, 1001613224]
            mck().get_request_id.side_effect = request_ids

            result = []
            async for x in bulkwalk('127.0.0.1', 'private', ['1.3.6.1.2.1.2.2'],
                                    bulk_size=20):
                result.append(x)

            mck.assert_called_with(timeout=6)
            assert mck().send.mock_calls == [
                call('127.0.0.1', 161, req1),
                call('127.0.0.1', 161, req2),
                call('127.0.0.1', 161, req3),
            ]

            expected = [
                VarBind('1.3.6.1.2.1.2.2.1.1.1', Integer(1)),
                VarBind('1.3.6.1.2.1.2.2.1.1.10', Integer(10)),
                VarBind('1.3.6.1.2.1.2.2.1.2.1', OctetString(b"lo")),
                VarBind('1.3.6.1.2.1.2.2.1.2.10', OctetString(b"eth0")),
                VarBind('1.3.6.1.2.1.2.2.1.3.1', Integer(24)),
                VarBind('1.3.6.1.2.1.2.2.1.3.10', Integer(6)),
                VarBind('1.3.6.1.2.1.2.2.1.4.1', Integer(65536)),
                VarBind('1.3.6.1.2.1.2.2.1.4.10', Integer(1500)),
                VarBind('1.3.6.1.2.1.2.2.1.5.1',  Gauge(10000000)),
                VarBind('1.3.6.1.2.1.2.2.1.5.10',  Gauge(4294967295)),
                VarBind('1.3.6.1.2.1.2.2.1.6.1', OctetString(b"")),
                VarBind('1.3.6.1.2.1.2.2.1.6.10',
                        OctetString(b"\x02\x42\xAC\x11\x00\x02")),
                VarBind('1.3.6.1.2.1.2.2.1.7.1', Integer(1)),
                VarBind('1.3.6.1.2.1.2.2.1.7.10', Integer(1)),
                VarBind('1.3.6.1.2.1.2.2.1.8.1', Integer(1)),
                VarBind('1.3.6.1.2.1.2.2.1.8.10', Integer(1)),
                VarBind('1.3.6.1.2.1.2.2.1.9.1', TimeTicks(0)),
                VarBind('1.3.6.1.2.1.2.2.1.9.10', TimeTicks(0)),
                VarBind('1.3.6.1.2.1.2.2.1.10.1', Counter(172)),
                VarBind('1.3.6.1.2.1.2.2.1.10.10', Counter(60558)),
                VarBind('1.3.6.1.2.1.2.2.1.11.1', Counter(2)),
                VarBind('1.3.6.1.2.1.2.2.1.11.10', Counter(564)),
                VarBind('1.3.6.1.2.1.2.2.1.12.1', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.12.10', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.13.1', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.13.10', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.14.1', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.14.10', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.15.1', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.15.10', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.16.1', Counter(172)),
                VarBind('1.3.6.1.2.1.2.2.1.16.10', Counter(44295)),
                VarBind('1.3.6.1.2.1.2.2.1.17.1', Counter(2)),
                VarBind('1.3.6.1.2.1.2.2.1.17.10', Counter(442)),
                VarBind('1.3.6.1.2.1.2.2.1.18.1', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.18.10', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.19.1', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.19.10', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.20.1', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.20.10', Counter(0)),
                VarBind('1.3.6.1.2.1.2.2.1.21.1',  Gauge(0)),
                VarBind('1.3.6.1.2.1.2.2.1.21.10',  Gauge(0)),
                VarBind('1.3.6.1.2.1.2.2.1.22.1', ObjectIdentifier(0, 0)),
                VarBind('1.3.6.1.2.1.2.2.1.22.10', ObjectIdentifier(0, 0))
            ]
        assert result == expected



# pylint: disable=redefined-builtin
# pylint: disable=invalid-name
"""
This module contains the high-level functions to access the library. Care is
taken to make this as pythonic as possible and hide as many of the gory
implementations as possible.
"""

from __future__ import unicode_literals

from collections import namedtuple

# !!! DO NOT REMOVE !!! The following import triggers the processing of SNMP
# Types and thus populates the Registry. If this is not included, Non x.690
# SNMP types will not be properly detected!
import puresnmp.types
from puresnmp.api.pythonic import (
    bulkget,
    bulktable,
    bulkwalk,
    get,
    getnext,
    multiget,
    multigetnext,
    multiset,
    multiwalk,
    set,
    table,
    traps,
    walk
)

try:
    import importlib.metadata as im
except ImportError:
    import importlib_metadata as im

__version__ = im.version("puresnmp")



# pylint: disable=missing-docstring
"""
Test the "external" interface in version 1.0.

This file should remain in the repository as long as we're still backwrds
compatible with 1.0. This includes call signatures *including* return types!
"""


import datetime
import sys
from collections import OrderedDict
from typing import TYPE_CHECKING

import six

import puresnmp as snmp
from puresnmp.pdu import VarBind
from puresnmp.util import BulkResult
from puresnmp.x690.types import ObjectIdentifier, OctetString
from puresnmp.const import Version

from .. import ByteTester, readbytes, readbytes_multiple

if TYPE_CHECKING:
    # pylint: disable=unused-import
    from typing import Type


if sys.version_info >= (3, 0):
    from unittest.mock import patch
else:
    from mock import patch

OID = ObjectIdentifier.from_string
TESTS_SHOULD_RUN = snmp.__version__.startswith('1')


def assert_of_types(values, types):
    '''
    Checks the types of a collection of values. Raises an AssertionError if
    something is of the wrong type.

    This assumes that *values* and *types* are collections of tuples. Example::

        >>> values = [(1, 2), (1, 'foo')]
        >>> expected_types = [(int, int), (int, str)]
        >>> assert_of_types(values, expected_types)
    '''
    for result, expected in zip(values, types):
        for a, b in zip(result, expected):
            if not isinstance(a, b):
                raise AssertionError('%r is not of type %r' % (a, b))


class TestGet(ByteTester):

    def setUp(self):
        patcher = patch('puresnmp.api.raw.Transport')
        Transport = patcher.start()
        self.transport = Transport()
        self.transport.get_request_id.return_value = 1
        self.addCleanup(lambda: patcher.stop)

    def test_get(self):
        response = readbytes('apiv1/get_response.hex')
        self.transport.send.return_value = response
        result = snmp.get('192.168.1.1', 'private', '1.3.6.1.2.1.1.2.0')
        self.assertEqual(result, '1.3.6.1.4.1.8072.3.2.10')
        self.assertTrue(self.transport.send.called, 'method was not called')
        if six.PY3:
            self.assertIsInstance(result, str)
        else:
            self.assertIsInstance(result, unicode)

    def test_multiget(self):
        response = readbytes('apiv1/multiget_response.hex')
        self.transport.send.return_value = response
        oids = ['1.3.6.1.2.1.1.4.0', '1.3.6.1.2.1.1.6.0']
        result = snmp.multiget('192.168.1.1', 'private', oids,
                               port=161, timeout=1)
        self.assertEqual(result, [b'root', b'On the move'])
        self.assertTrue(self.transport.send.called, 'method was not called')
        for value in result:
            self.assertIsInstance(value, bytes)

    def test_multiget_v1(self):
        response = readbytes('apiv1/multiget_response.hex')
        self.transport.send.return_value = response
        oids = ['1.3.6.1.2.1.1.4.0', '1.3.6.1.2.1.1.6.0']
        result = snmp.multiget('192.168.1.1', 'private', oids,
                               port=161, timeout=1, version=Version.V1)
        self.assertEqual(result, [b'root', b'On the move'])
        self.assertTrue(self.transport.send.called, 'method was not called')
        for value in result:
            self.assertIsInstance(value, bytes)


    def test_getnext(self):
        response = readbytes('apiv1/getnext_response.hex')
        self.transport.send.return_value = response
        result = snmp.getnext('192.168.1.1', 'private', '1.3.6.1.2.1.1.6.0',
                              port=161, timeout=1)
        expected = VarBind(OID('1.3.6.1.2.1.1.7.0'), 72)
        self.assertEqual(result, expected)
        self.assertTrue(self.transport.send.called, 'method was not called')
        self.assertIsInstance(result.oid, ObjectIdentifier)
        self.assertIsInstance(result.value, int)

    def test_bulkget(self):
        response = readbytes('apiv1/bulkget_response.hex')
        self.transport.send.return_value = response
        result = snmp.bulkget('192.168.1.1', 'private',
                              scalar_oids=['1.3.6.1.2.1.1.1.0'],
                              repeating_oids=[
                                  '1.3.6.1.2.1.3.1.1',
                                  '1.3.6.1.2.1.4',
                              ],
                              max_list_size=10,
                              port=161, timeout=1)

        expected = BulkResult(
            {'1.3.6.1.2.1.1.2.0': '1.3.6.1.4.1.8072.3.2.10'},
            OrderedDict([
                ('1.3.6.1.2.1.3.1.1.1.12.1.172.17.0.1', 12),
                ('1.3.6.1.2.1.4.1.0', 1),
                ('1.3.6.1.2.1.3.1.1.2.12.1.172.17.0.1', b'\x02B@j\xbf\xcd'),
                ('1.3.6.1.2.1.4.2.0', 64),
                ('1.3.6.1.2.1.3.1.1.3.12.1.172.17.0.1', b'\xac\x11\x00\x01'),
                ('1.3.6.1.2.1.4.3.0', 589),
                ('1.3.6.1.2.1.4.4.0', 0),
                ('1.3.6.1.2.1.4.5.0', 0),
                ('1.3.6.1.2.1.4.6.0', 0),
                ('1.3.6.1.2.1.4.7.0', 0),
                ('1.3.6.1.2.1.4.8.0', 0),
                ('1.3.6.1.2.1.4.9.0', 410),
                ('1.3.6.1.2.1.4.10.0', 409)
            ])
        )
        self.assertEqual(result, expected)
        self.assertTrue(self.transport.send.called, 'method was not called')
        key_types = {type(k) for k in result.listing.keys()}

        if six.PY3:
            self.assertIsInstance(result.scalars['1.3.6.1.2.1.1.2.0'], str)
            self.assertEqual(key_types, {str})
        else:
            self.assertIsInstance(result.scalars['1.3.6.1.2.1.1.2.0'], unicode)
            self.assertEqual(key_types, {unicode})

        expected_types = [
            int,
            int,
            bytes,
            int,
            bytes,
            int,
            int,
            int,
            int,
            int,
            int,
            int,
            int
        ]
        for value, type_ in zip(result.listing.values(), expected_types):
            self.assertIsInstance(value, type_)

    def test_bulkget_v1(self):
        response = readbytes('apiv1/bulkget_response.hex')
        self.transport.send.return_value = response
        result = snmp.bulkget('192.168.1.1', 'private',
                              scalar_oids=['1.3.6.1.2.1.1.1.0'],
                              repeating_oids=[
                                  '1.3.6.1.2.1.3.1.1',
                                  '1.3.6.1.2.1.4',
                              ],
                              max_list_size=10,
                              port=161, timeout=1, version=Version.V1)

        expected = BulkResult(
            {'1.3.6.1.2.1.1.2.0': '1.3.6.1.4.1.8072.3.2.10'},
            OrderedDict([
                ('1.3.6.1.2.1.3.1.1.1.12.1.172.17.0.1', 12),
                ('1.3.6.1.2.1.4.1.0', 1),
                ('1.3.6.1.2.1.3.1.1.2.12.1.172.17.0.1', b'\x02B@j\xbf\xcd'),
                ('1.3.6.1.2.1.4.2.0', 64),
                ('1.3.6.1.2.1.3.1.1.3.12.1.172.17.0.1', b'\xac\x11\x00\x01'),
                ('1.3.6.1.2.1.4.3.0', 589),
                ('1.3.6.1.2.1.4.4.0', 0),
                ('1.3.6.1.2.1.4.5.0', 0),
                ('1.3.6.1.2.1.4.6.0', 0),
                ('1.3.6.1.2.1.4.7.0', 0),
                ('1.3.6.1.2.1.4.8.0', 0),
                ('1.3.6.1.2.1.4.9.0', 410),
                ('1.3.6.1.2.1.4.10.0', 409)
            ])
        )
        self.assertEqual(result, expected)
        self.assertTrue(self.transport.send.called, 'method was not called')
        key_types = {type(k) for k in result.listing.keys()}

        if six.PY3:
            self.assertIsInstance(result.scalars['1.3.6.1.2.1.1.2.0'], str)
            self.assertEqual(key_types, {str})
        else:
            self.assertIsInstance(result.scalars['1.3.6.1.2.1.1.2.0'], unicode)
            self.assertEqual(key_types, {unicode})

        expected_types = [
            int,
            int,
            bytes,
            int,
            bytes,
            int,
            int,
            int,
            int,
            int,
            int,
            int,
            int
        ]
        for value, type_ in zip(result.listing.values(), expected_types):
            self.assertIsInstance(value, type_)

    def test_bulkwalk(self):
        response = readbytes_multiple('apiv1/bulkwalk_response.hex')
        self.transport.send.side_effect = response
        result = list(snmp.bulkwalk('192.168.1.1', 'private', [
            '1.3.6.1.2.1.1.9.1.4']))

        expected = [
            VarBind(OID('1.3.6.1.2.1.1.9.1.4.1'), datetime.timedelta(0)),
            VarBind(OID('1.3.6.1.2.1.1.9.1.4.2'), datetime.timedelta(0)),
            VarBind(OID('1.3.6.1.2.1.1.9.1.4.3'), datetime.timedelta(0)),
            VarBind(OID('1.3.6.1.2.1.1.9.1.4.4'), datetime.timedelta(0)),
            VarBind(OID('1.3.6.1.2.1.1.9.1.4.5'), datetime.timedelta(0)),
            VarBind(OID('1.3.6.1.2.1.1.9.1.4.6'), datetime.timedelta(0)),
            VarBind(OID('1.3.6.1.2.1.1.9.1.4.7'), datetime.timedelta(0)),
            VarBind(OID('1.3.6.1.2.1.1.9.1.4.8'), datetime.timedelta(0)),
            VarBind(OID('1.3.6.1.2.1.1.9.1.4.9'), datetime.timedelta(0)),
            VarBind(OID('1.3.6.1.2.1.1.9.1.4.10'), datetime.timedelta(0))
        ]

        self.assertEqual(result, expected)
        self.assertTrue(self.transport.send.called, 'method was not called')
        expected_types = [(ObjectIdentifier, datetime.timedelta)] * 10
        returned_values = [(row.oid, row.value) for row in result]
        assert_of_types(returned_values, expected_types)

    def test_multigetnext(self):
        response = readbytes('apiv1/multigetnext_response.hex')
        self.transport.send.return_value = response
        result = snmp.multigetnext('192.168.1.1', 'private',
                                   ['1.3.6.1.2.1.3.1.1', '1.3.6.1.2.1.4'])
        expected = [
            VarBind(OID('1.3.6.1.2.1.3.1.1.1.12.1.172.17.0.1'), 12),
            VarBind(OID('1.3.6.1.2.1.4.1.0'), 1)
        ]
        self.assertEqual(result, expected)
        self.assertTrue(self.transport.send.called, 'method was not called')
        expected_types = [
            (ObjectIdentifier, int),
            (ObjectIdentifier, int),
        ]
        returned_values = [(row.oid, row.value) for row in result]
        returned_types = [(type(row.oid), type(row.value)) for row in result]
        assert_of_types(returned_values, expected_types)

    def test_multigetnext_v1(self):
        response = readbytes('apiv1/multigetnext_response.hex')
        self.transport.send.return_value = response
        result = snmp.multigetnext('192.168.1.1', 'private',
                                   ['1.3.6.1.2.1.3.1.1', '1.3.6.1.2.1.4'], version=Version.V1)
        expected = [
            VarBind(OID('1.3.6.1.2.1.3.1.1.1.12.1.172.17.0.1'), 12),
            VarBind(OID('1.3.6.1.2.1.4.1.0'), 1)
        ]
        self.assertEqual(result, expected)
        self.assertTrue(self.transport.send.called, 'method was not called')
        expected_types = [
            (ObjectIdentifier, int),
            (ObjectIdentifier, int),
        ]
        returned_values = [(row.oid, row.value) for row in result]
        returned_types = [(type(row.oid), type(row.value)) for row in result]
        assert_of_types(returned_values, expected_types)

    def test_multiset(self):
        response = readbytes('apiv1/multiset_response.hex')
        self.transport.send.return_value = response
        result = snmp.multiset('127.0.0.1', 'private', [
            ('1.3.6.1.2.1.1.4.0', OctetString(b'foo')),
            ('1.3.6.1.2.1.1.6.0', OctetString(b'bar'))
        ])
        expected = {
            '1.3.6.1.2.1.1.4.0': b'foo',
            '1.3.6.1.2.1.1.6.0': b'bar'
        }
        self.assertEqual(result, expected)
        self.assertTrue(self.transport.send.called, 'method was not called')
        self.assertIsInstance(result, dict)
        key_types = {type(key) for key in result.keys()}
        if six.PY3:
            self.assertEqual(key_types, {str})
        else:
            self.assertEqual(key_types, {unicode})
        value_types = {type(value) for value in result.values()}
        self.assertEqual(value_types, {bytes})

    def test_multiwalk(self):
        response_1 = readbytes('apiv1/multiwalk_response_1.hex')
        response_2 = readbytes('apiv1/multiwalk_response_2.hex')
        response_3 = readbytes('apiv1/multiwalk_response_3.hex')

        self.transport.send.side_effect = [
            response_1,
            response_2,
            response_3,
        ]
        result = snmp.multiwalk('127.0.0.1', 'private', [
            '1.3.6.1.2.1.2.2.1.1',
            '1.3.6.1.2.1.1.2.1',
        ])
        expected = [
            VarBind(OID('1.3.6.1.2.1.2.2.1.1.1'), 1),
            VarBind(OID('1.3.6.1.2.1.2.2.1.1.6'), 6),
        ]
        result = list(result)
        self.assertEqual(result, expected)

        expected_types = [
            (ObjectIdentifier, int),
            (ObjectIdentifier, int),
        ]
        returned_values = [(row.oid, row.value) for row in result]
        assert_of_types(returned_values, expected_types)

    def test_multiwalk_v1(self):
        response_1 = readbytes('apiv1/multiwalk_response_1.hex')
        response_2 = readbytes('apiv1/multiwalk_response_2.hex')
        response_3 = readbytes('apiv1/multiwalk_response_3.hex')

        self.transport.send.side_effect = [
            response_1,
            response_2,
            response_3,
        ]
        result = snmp.multiwalk('127.0.0.1', 'private', [
            '1.3.6.1.2.1.2.2.1.1',
            '1.3.6.1.2.1.1.2.1',
        ], version=Version.V1)
        expected = [
            VarBind(OID('1.3.6.1.2.1.2.2.1.1.1'), 1),
            VarBind(OID('1.3.6.1.2.1.2.2.1.1.6'), 6),
        ]
        result = list(result)
        self.assertEqual(result, expected)

        expected_types = [
            (ObjectIdentifier, int),
            (ObjectIdentifier, int),
        ]
        returned_values = [(row.oid, row.value) for row in result]
        assert_of_types(returned_values, expected_types)

    def test_set(self):
        self.transport.send.return_value = readbytes('apiv1/set_response.hex')
        result = snmp.set(
            '127.0.0.1',
            'private',
            '1.3.6.1.2.1.1.6.0',
            OctetString(b'Hello')
        )
        self.assertEqual(result, b'On the move')
        self.assertTrue(self.transport.send.called, 'method was not called')
        self.assertIsInstance(result, bytes)

    def test_table(self):
        responses = readbytes_multiple('apiv1/table_response.hex')
        self.transport.send.side_effect = responses
        result = snmp.table('127.0.0.1', 'private', '1.3.6.1.2.1.2.2')

        # Sort the result for Python < 3.6 (unsorted dicts)
        result = sorted(result, key=lambda item: item['0'])
        expected = [{
            '0': '1',
            '1': 1,
            '10': 172,
            '11': 2,
            '12': 0,
            '13': 0,
            '14': 0,
            '15': 0,
            '16': 172,
            '17': 2,
            '18': 0,
            '19': 0,
            '2': b'lo',
            '20': 0,
            '21': 0,
            '22': '0.0',
            '3': 24,
            '4': 65536,
            '5': 10000000,
            '6': b'',
            '7': 1,
            '8': 1,
            '9': datetime.timedelta(0)
        }, {
            '0': '12',
            '1': 12,
            '10': 13952,
            '11': 140,
            '12': 0,
            '13': 0,
            '14': 0,
            '15': 0,
            '16': 4391,
            '17': 51,
            '18': 0,
            '19': 0,
            '2': b'eth0',
            '20': 0,
            '21': 0,
            '22': '0.0',
            '3': 6,
            '4': 1500,
            '5': 4294967295,
            '6': b'\x02B\xac\x11\x00\x02',
            '7': 1,
            '8': 1,
            '9': datetime.timedelta(0)
        }]
        self.assertEqual(result, expected)
        self.assertEqual(self.transport.send.call_count, 45)
        for row in result:
            self.assertIsInstance(row, dict)
            dict_types = {type(key) for key in row.keys()}
            if six.PY3:
                self.assertEqual(dict_types, {str})
            else:
                self.assertEqual(dict_types, {unicode})

    def test_walk(self):
        responses = readbytes_multiple('apiv1/walk_response.hex')
        self.transport.send.side_effect = responses
        result = snmp.walk('127.0.0.1', 'private', '1.3.6.1.2.1.2.2.1.1')
        result = list(result)
        expected = [
            VarBind(oid=OID('1.3.6.1.2.1.2.2.1.1.1'), value=1),
            VarBind(oid=OID('1.3.6.1.2.1.2.2.1.1.12'), value=12)
        ]
        six.assertCountEqual(self, result, expected)
        self.assertEqual(self.transport.send.call_count, 3)

        expected_types = [
            (ObjectIdentifier, int),
            (ObjectIdentifier, int),
        ]
        returned_values = [
            (row.oid, row.value)
            for row in result
        ]
        assert_of_types(returned_values, expected_types)



from puresnmp.pdu import VarBind
from puresnmp.util import WalkRow, get_unfinished_walk_oids, group_varbinds
from puresnmp.x690.types import Null, ObjectIdentifier

OID = ObjectIdentifier.from_string


def test_group_varbinds():
    '''
    "group_varbinds" should convert an interleaved list of OIDs into a more
    usable dictionary.
    '''
    varbinds = [
        VarBind(OID('1.1.1'), Null()),
        VarBind(OID('2.2.1'), Null()),
        VarBind(OID('3.3.1'), Null()),
        VarBind(OID('1.1.2'), Null()),
        VarBind(OID('2.2.2'), Null()),
        VarBind(OID('3.3.2'), Null()),
        VarBind(OID('1.1.3'), Null()),
        VarBind(OID('2.2.3'), Null()),
        VarBind(OID('3.3.3'), Null()),
    ]
    effective_roots = [
        OID('1.1'),
        OID('2.2'),
        OID('3.3'),
    ]
    result = group_varbinds(varbinds, effective_roots)
    expected = {
        OID('1.1'): [
            VarBind(OID('1.1.1'), Null()),
            VarBind(OID('1.1.2'), Null()),
            VarBind(OID('1.1.3'), Null()),
        ],
        OID('2.2'): [
            VarBind(OID('2.2.1'), Null()),
            VarBind(OID('2.2.2'), Null()),
            VarBind(OID('2.2.3'), Null()),
        ],
        OID('3.3'): [
            VarBind(OID('3.3.1'), Null()),
            VarBind(OID('3.3.2'), Null()),
            VarBind(OID('3.3.3'), Null()),
        ],
    }

    assert result == expected


def test_get_unfinished_walk_oids():
    oid_groups = {
        OID('1.1'): [
            VarBind(OID('1.1.1'), Null()),
            VarBind(OID('1.1.2'), Null()),
        ],
        OID('2.2'): [
            VarBind(OID('2.2.1'), Null()),
            VarBind(OID('2.2.2'), Null()),
        ],
        OID('3.3'): [
            VarBind(OID('3.3.1'), Null()),
            VarBind(OID('3.4.2'), Null()),
        ]
    }
    result = get_unfinished_walk_oids(oid_groups)
    expected = [
        (OID('1.1'), WalkRow(VarBind(OID('1.1.2'), Null()), unfinished=True)),
        (OID('2.2'), WalkRow(VarBind(OID('2.2.2'), Null()), unfinished=True)),
    ]
    assert result == expected



"""
Model for SNMP PDUs (Request/Response messages).

PDUs all have a common structure, which is handled in the
:py:class:`~.PDU` class. The different (basic) PDU types only differ in
their type identifier header (f.ex. ``b'\\xa0'`` for a
:py:class:`~.GetRequest`).
"""

# TODO: Add a method to wrap a message in a full packet (including SNMP version
#       and community). This can then replace some duplicated code in
#       "puresnmp.get", "puresnmp.walk" & co.

from typing import TYPE_CHECKING, Iterable, Tuple, cast

import six

from .const import MAX_VARBINDS
from .exc import EmptyMessage, ErrorResponse, NoSuchOID, TooManyVarbinds
from .snmp import ERROR_MESSAGES, VarBind
from .typevars import PyType, SocketInfo
from .x690.types import (
    Integer,
    Null,
    ObjectIdentifier,
    Sequence,
    Type,
    pop_tlv
)
from .x690.util import TypeInfo, encode_length, to_bytes

if TYPE_CHECKING:
    # pylint: disable=unused-import
    from typing import Any, Iterator, List, Union, Optional


if six.PY3:
    unicode = str  # pylint: disable=invalid-name


class PDU(Type):  # type: ignore
    """
    The superclass for SNMP Messages (GET, SET, GETNEXT, ...)
    """
    TYPECLASS = TypeInfo.CONTEXT
    TAG = 0

    @classmethod
    def decode(cls, data):
        # type: (bytes) -> PDU
        """
        This method takes a :py:class:`bytes` object and converts it to
        an application object. This is callable from each subclass of
        :py:class:`~.PDU`.
        """
        # TODO (advanced): recent tests revealed that this is *not symmetric*
        # with __bytes__ of this class. This should be ensured!
        if not data:
            raise EmptyMessage('No data to decode!')
        request_id, data = pop_tlv(data)
        error_status, data = pop_tlv(data)
        error_index, data = pop_tlv(data)
        if error_status.value:
            error_detail, data = cast(
                Tuple[Iterable[Tuple[ObjectIdentifier, int]], bytes],
                pop_tlv(data)
            )
            if not isinstance(error_detail, Sequence):
                raise TypeError(
                    'error-detail should be a sequence but got %r' %
                    type(error_detail))
            varbinds = [VarBind(*raw_varbind) for raw_varbind in error_detail]
            if error_index.value != 0:
                offending_oid = varbinds[error_index.value-1].oid
            else:
                # Offending OID is unknown
                offending_oid = None
            assert data == b''
            exception = ErrorResponse.construct(
                error_status.value, offending_oid)
            raise exception

        values, data = cast(
            Tuple[Iterable[Tuple[ObjectIdentifier, int]], bytes],
            pop_tlv(data)
        )
        if not isinstance(values, Sequence):
            raise TypeError('PDUs can only be decoded from sequences but got '
                            '%r instead' % type(values))

        varbinds = []
        for oid, value in values:
            # NOTE: this uses the "is" check to make 100% sure we check against
            # the sentinel object defined in this module!
            if value is END_OF_MIB_VIEW:
                varbinds.append(VarBind(oid, END_OF_MIB_VIEW))
                break
            varbinds.append(VarBind(oid, value))

        return cls(
            request_id.value,
            varbinds,
            error_status.value,
            error_index.value
        )

    def __init__(self, request_id, varbinds, error_status=0, error_index=0):
        # type: (int, Union[VarBind, List[VarBind]], int, int) -> None
        self.request_id = request_id
        self.error_status = error_status
        self.error_index = error_index
        if isinstance(varbinds, VarBind):
            self.varbinds = [VarBind(*varbinds)]
        else:
            self.varbinds = varbinds

    def __bytes__(self):
        # type: () -> bytes
        wrapped_varbinds = [Sequence(vb.oid, vb.value)  # type: ignore
                            for vb in self.varbinds]
        data = [
            Integer(self.request_id),
            Integer(self.error_status),
            Integer(self.error_index),
            Sequence(*wrapped_varbinds)
        ]
        payload = b''.join([to_bytes(chunk) for chunk in data])

        tinfo = TypeInfo(TypeInfo.CONTEXT, TypeInfo.CONSTRUCTED, self.TAG)
        length = encode_length(len(payload))
        return to_bytes(tinfo) + length + payload

    def __repr__(self):
        # type: () -> str
        return '%s(%r, %r)' % (
            self.__class__.__name__,
            self.request_id, self.varbinds)

    def __eq__(self, other):
        # type: (Any) -> bool
        # pylint: disable=unidiomatic-typecheck
        return (type(other) == type(self) and
                self.request_id == other.request_id and
                self.varbinds == other.varbinds)

    def pretty(self):  # pragma: no cover
        # type: () -> str
        """
        Returns a "prettified" string representing the SNMP message.
        """
        lines = [
            self.__class__.__name__,
            '    Request ID: %s' % self.request_id,
            '    Error Status: %s' % self.error_status,
            '    Error Index: %s' % self.error_index,
            '    Varbinds: ',
        ]
        for bind in self.varbinds:
            lines.append('        %s: %s' % (bind.oid, bind.value))  # type: ignore

        return '\n'.join(lines)


class EndOfMibView(PDU):
    """
    Sentinel value to detect endOfMibView
    """
    # This subclassesPDU for type-consistency

    def __init__(self):
        # type: () -> None
        super(EndOfMibView, self).__init__(-1, [], 0, 0)


#: Singleton instance of "EndOfMibView"
END_OF_MIB_VIEW = EndOfMibView()


class GetRequest(PDU):
    """
    Represents an SNMP Get Request.
    """
    TAG = 0

    def __init__(self, request_id, *oids):
        # type: (int, Union[str, ObjectIdentifier]) -> None
        if len(oids) > MAX_VARBINDS:
            raise TooManyVarbinds(len(oids))
        wrapped_oids = []
        for oid in oids:
            if isinstance(oid, str):
                wrapped_oids.append(ObjectIdentifier.from_string(oid))
            else:
                wrapped_oids.append(oid)
        super(GetRequest, self).__init__(
            request_id,
            [VarBind(oid, Null()) for oid in wrapped_oids])  # type: ignore


class GetResponse(PDU):
    """
    Represents an SNMP basic response (this may be returned for other requests
    than GET as well).
    """
    TAG = 2

    @classmethod
    def decode(cls, data):
        # type: (bytes) -> PDU
        """
        Try decoding the response. If nothing was returned (the message was
        empty), raise a :py:exc:`~puresnmp.exc.NoSuchOID` exception.
        """
        # TODO This is not reslly clean, but it should work. A GetResponse has
        #      the same type identifier as a "endOfMibView" *value*. The way
        #      puresnmp is structured (recursively calling "pop_tlv") makes it
        #      difficult to distinguish between a valid GetResponse object and
        #      an endOfMibView value, except that the endOfMibView had no data.
        if not data:
            return END_OF_MIB_VIEW
        try:
            return super(GetResponse, cls).decode(data)
        except EmptyMessage as exc:
            raise NoSuchOID(
                ObjectIdentifier(0),
                'Nothing found at the given OID (%s)' % exc)


class GetNextRequest(GetRequest):
    """
    Represents an SNMP GetNext Request.
    """
    TAG = 1


class SetRequest(PDU):
    """
    Represents an SNMP SET Request.
    """
    TAG = 3


class BulkGetRequest(Type):  # type: ignore
    """
    Represents a SNMP GetBulk request
    """
    TYPECLASS = TypeInfo.CONTEXT
    TAG = 5

    def __init__(self, request_id, non_repeaters, max_repeaters, *oids):
        # type: (int, int, int, ObjectIdentifier) -> None
        if len(oids) > MAX_VARBINDS:
            raise TooManyVarbinds(len(oids))
        self.request_id = request_id
        self.non_repeaters = non_repeaters
        self.max_repeaters = max_repeaters
        self.varbinds = []  # type: List[VarBind]
        for oid in oids:
            self.varbinds.append(VarBind(oid, Null()))  # type: ignore

    def __bytes__(self):
        # type: () -> bytes
        wrapped_varbinds = [Sequence(vb.oid, vb.value)  # type: ignore
                            for vb in self.varbinds]
        data = [
            Integer(self.request_id),
            Integer(self.non_repeaters),
            Integer(self.max_repeaters),
            Sequence(*wrapped_varbinds)
        ]
        payload = b''.join([to_bytes(chunk) for chunk in data])

        tinfo = TypeInfo(TypeInfo.CONTEXT, TypeInfo.CONSTRUCTED, self.TAG)
        length = encode_length(len(payload))
        return to_bytes(tinfo) + length + payload

    def __repr__(self):
        # type: () -> str
        oids = [repr(oid) for oid, _ in self.varbinds]
        return '%s(%r, %r, %r, %s)' % (
            self.__class__.__name__,
            self.request_id,
            self.non_repeaters,
            self.max_repeaters,
            ', '.join(oids))

    def __eq__(self, other):
        # type: (Any) -> bool
        # pylint: disable=unidiomatic-typecheck
        return (type(other) == type(self) and
                self.request_id == other.request_id and
                self.non_repeaters == other.non_repeaters and
                self.max_repeaters == other.max_repeaters and
                self.varbinds == other.varbinds)

    def pretty(self):  # pragma: no cover
        # type: () -> str
        """
        Returns a "prettified" string representing the SNMP message.
        """
        lines = [
            self.__class__.__name__,
            '    Request ID: %s' % self.request_id,
            '    Non Repeaters: %s' % self.non_repeaters,
            '    Max Repeaters: %s' % self.max_repeaters,
            '    Varbinds: ',
        ]
        for bind in self.varbinds:
            lines.append('        %s: %s' % (bind.oid, bind.value))  # type: ignore

        return '\n'.join(lines)


class InformRequest(PDU):
    """
    Represents an SNMP Inform request
    """
    TAG = 6


class Trap(PDU):
    """
    Represents an SNMP Trap
    """

    TAG = 7

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        super(Trap, self).__init__(*args, **kwargs)
        self.source = None  # type: Optional[SocketInfo]



"""
SMI Types / Structure types which are not defined in :term:`X.690`.

See `RFC 1155 section 3.2.3`_ for a description of the types and `RFC 3416`_
for the definition of the new types.

.. note::
    The IPv6 Type is not yet implemented and will be returned as OctetString!

.. _RFC 1155 section 3.2.3: https://tools.ietf.org/html/rfc1155#section-3.2.3
.. _RFC 3416: https://tools.ietf.org/html/rfc3416
"""
# TODO: Implement IPv6 via https://tools.ietf.org/html/rfc2465

import sys
from datetime import timedelta
from ipaddress import IPv4Address
from struct import pack
from typing import TYPE_CHECKING, Union

from .x690.types import Integer, OctetString
from .x690.util import TypeInfo

if TYPE_CHECKING:
    from typing import Optional


class IpAddress(OctetString):
    """
    SNMP Type for IPv4 Addresses
    """
    TYPECLASS = TypeInfo.APPLICATION
    TAG = 0x00

    def __init__(self, value):
        # type: (bytes) -> None
        if isinstance(value, IPv4Address):
            remainder = int(value)
            octet_4, remainder = remainder & 0xff, remainder >> 8
            octet_3, remainder = remainder & 0xff, remainder >> 8
            octet_2, remainder = remainder & 0xff, remainder >> 8
            octet_1, remainder = remainder & 0xff, remainder >> 8
            value = pack('BBBB', octet_1, octet_2, octet_3, octet_4)
        super(IpAddress, self).__init__(value)

    def pythonize(self):
        # type: () -> Optional[bytes]
        return self.value

        # TODO The following code breaks backwards compatbility and should be
        # released in the next mator verion

        # TODO v2.0.0 intvalue = 0
        # TODO v2.0.0 for i, octet in enumerate(reversed(self.value)):
        # TODO v2.0.0     if sys.version_info < (3, 0):
        # TODO v2.0.0         # Python 2 assumes has str === bytes so we need to cast
        # TODO v2.0.0         octet = ord(octet)
        # TODO v2.0.0     intvalue |= octet << (8*i)
        # TODO v2.0.0 return ip_address(intvalue)


class Counter(Integer):
    """
    SNMP type for counters.
    """
    SIGNED = False
    TYPECLASS = TypeInfo.APPLICATION
    TAG = 0x01

    def __init__(self, value):
        # type: (int) -> None
        value &= 0xFFFFFFFF if value >= 2 ** 32 else value
        if value <= 0:
            value = 0
        super(Counter, self).__init__(value)


class Gauge(Integer):
    """
    SNMP type for gauges.
    """
    SIGNED = False
    TYPECLASS = TypeInfo.APPLICATION
    TAG = 0x02


class TimeTicks(Integer):
    """
    SNMP type for time ticks.
    """
    SIGNED = False
    TYPECLASS = TypeInfo.APPLICATION
    TAG = 0x03

    def __init__(self, value):
        # type: (Union[timedelta, int]) -> None
        if isinstance(value, timedelta):
            value_int = int(value.total_seconds() * 100)
        else:
            value_int = value
        super(TimeTicks, self).__init__(value_int)

    def pythonize(self):
        # type: () -> Optional[timedelta]
        if self.value is None:
            return None
        seconds = self.value / 100.0  # see rfc2578#section-7.1.8
        return timedelta(seconds=seconds)


class Opaque(OctetString):
    TYPECLASS = TypeInfo.APPLICATION
    TAG = 0x04


class NsapAddress(Integer):
    TYPECLASS = TypeInfo.APPLICATION
    TAG = 0x05


class Counter64(Integer):
    """
    As defined in RFC 2578
    """
    SIGNED = False
    TYPECLASS = TypeInfo.APPLICATION
    TAG = 0x06

    def __init__(self, value):
        # type: (int) -> None
        value &= 0xFFFFFFFFFFFFFFFF if value >= 2 ** 64 else value
        if value <= 0:
            value = 0
        super(Counter64, self).__init__(value)


def _walk_subclasses(cls, indent=0):  # pragma: no cover
    # type: (type, int) -> None
    '''
    Recursively walk over the :py:class:`Type` hierarchy and print out ReST
    formatted text on stdout.
    '''
    if cls.__module__ == '__main__':
        modname = 'puresnmp.types'
    else:
        modname = cls.__module__

    cname = '.'.join([modname, cls.__qualname__])
    ref = ':py:class:`%s`' % cname

    print('\n', '   ' * indent, '* ', ref)
    for subclass in sorted(cls.__subclasses__(),
                           key=lambda x: x.__module__ + x.__name__):
        _walk_subclasses(subclass, indent + 1)


def main():  # pragma: no cover
    # type: () -> int
    """
    Entrypoint for::

        python -m puresnmp.types

    This will output a RST formatted document containing the available types.
    This function was written to generate a documentation page with the
    available types.
    """
    from .x690.types import Type
    print('.. _type_tree:\n')
    print('Type Tree')
    print('=========\n')
    _walk_subclasses(Type)
    return 0


if __name__ == '__main__':  # pragma: no cover
    import sys
    sys.exit(main())



# pylint: skip-file
from __future__ import print_function

import re
import sys
import unittest
from logging import Handler, LogRecord, getLevelName
from os.path import dirname, join
from typing import TYPE_CHECKING

import six

from ..x690.util import to_bytes

if TYPE_CHECKING:
    from typing import Any, Callable, Generator, List, Tuple, TypeVar, Union
    T = TypeVar('T')

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest  # type: ignore


DATA_DIR = join(dirname(__file__), 'data')

__unittest = True  # <- This disables stack traces in unittest output for
# everything in this module.


class CapturingHandler(Handler):

    def __init__(self):
        # type: () -> None
        super(CapturingHandler, self).__init__()
        self.captured_records = []  # type: List[LogRecord]

    def emit(self, record):
        # type: (LogRecord) -> None
        self.captured_records.append(record)

    def assertContains(self, level, message_regex):
        # type: (int, str) -> None
        found = False
        for record in self.captured_records:
            matches_level = record.levelno == level
            matches_re = re.search(message_regex, record.msg % record.args)
            if matches_level and matches_re:
                found = True
                break
        if not found:
            print('--- Captured log messages:', file=sys.stderr)
            for record in self.captured_records:
                print('Level:', getLevelName(record.levelno), 'Message:',
                      record.msg % record.args, file=sys.stderr)
            raise AssertionError('Pattern %r was not found with level %r in '
                                 'the log records' % (message_regex, level))


class ByteTester(unittest.TestCase):
    def assertBytesEqual(self, a, b):
        # type: (Union[bytes, bytearray], Union[bytes, bytearray]) -> None
        """
        Helper method to compare bytes with more helpful output.
        """
        def is_bytes(x):
            # type: (Union[bytes, bytearray]) -> bool
            return isinstance(x, (bytes, bytearray))
        if not is_bytes(a) or not is_bytes(b):
            raise ValueError('assertBytesEqual requires two bytes objects!')

        if a != b:
            comparisons = []
            type_a = type(a)
            type_b = type(b)
            a = bytearray(a)
            b = bytearray(b)

            def char_repr(c):
                # type: (bytes) -> str
                if 0x1f < char_a < 0x80:
                    # bytearray to prevent accidental pre-mature str conv
                    # str to prevent b'' suffix in repr's output
                    return repr(str(bytearray([char_a]).decode('ascii')))
                return '.'
            for offset, (char_a, char_b) in enumerate(zip_longest(a, b)):
                comp, marker = ('==', '') if char_a == char_b else ('!=', '>>')

                # Using "zip_longest", overflows are marked as "None", which is
                # unambiguous in this case, but we need to handle these
                # separately from the main format string.
                if char_a is None:
                    char_ab = char_ad = char_ah = char_ar = '?'
                else:
                    char_ab = '0b{:08b}'.format(char_a)
                    char_ad = '{:3d}'.format(char_a)
                    char_ah = '0x{:02x}'.format(char_a)
                    char_ar = char_repr(char_a)
                if char_b is None:
                    char_bb = char_bd = char_bh = char_br = '?'
                else:
                    char_bb = '0b{:08b}'.format(char_b)
                    char_bd = '{:3d}'.format(char_b)
                    char_bh = '0x{:02x}'.format(char_b)
                    char_br = char_repr(char_b)
                comparisons.append(
                    "{8:<3} Offset {0:4d}: "
                    "{1:^10} {4} {5:^10} | "
                    "{2:>3} {4} {6:>3} | "
                    "{3:^4} {4} {7:^4} | {9:>3} {4} {10:>3}".format(
                        offset,
                        char_ab,
                        char_ad,
                        char_ah,
                        comp,
                        char_bb,
                        char_bd,
                        char_bh,
                        marker,
                        char_ar,
                        char_br))
            raise AssertionError(
                'Bytes differ!\n' +
                'type(a)=%s, type(b)=%s\n' % (type_a, type_b) +
                '\nIndividual bytes:\n' +
                '\n'.join(comparisons))


def parse_meta_lines(lines):
    # type: (List[str]) -> Generator[Tuple[str, Any], None, None]
    pattern = re.compile(r'-\*-(.*?)-\*-')
    for line in lines:
        match = pattern.search(line)
        if not match:
            continue
        values = match.groups()[0]
        key, _, value = values.partition(':')
        if key.strip() == 'ascii-cols':
            start, _, end = value.partition('-')
            yield key.strip(), (int(start), int(end))
        else:
            yield key.strip(), value.strip()


def chunker(lines, is_boundary):
    # type: (List[T], Callable[[T], bool]) -> Generator[List[T], None, None]
    '''
    Transforms a list of items into a generator of new lists of items of the
    same type by looking for special boundary lines. Boundary lines are
    detected with the help of an "is_boundary" callable.

    Example:

        >>> data = [1, 2, -1, 3, 4]
        >>> for chunk in chunker(data, is_boundary=lambda x: x == -1):
        ...     print(chunk)
        [1, 2]
        [3, 4]
    '''
    if not lines:
        return

    collected = []  # type: List[T]
    for item in lines:
        if is_boundary(item):
            if collected:
                yield collected
            collected = []
            continue
        else:
            collected.append(item)
    if collected:
        yield collected


def detect_ascii_slice(lines):
    # type: (List[str]) -> slice
    '''
    Given a list of strings, this will return the most likely positions of byte
    positions. They are returned slice which should be able to extract the
    columns from each line.
    '''
    for line in lines:
        # if the content contains a ":" character, it contains the byte offset
        # in the beginning. This is the case for libsnmp command output using
        # the "-d" switch. We need to remove the offset
        match = re.match(r'^\d{4}:', line)
        if ':' in line:
            return slice(6, 56)
        else:
            return slice(0, 50)
    return slice(0, -1)


def readbytes_multiple(filename, base_dir=DATA_DIR):
    # type: (str, str) -> Generator[bytes, None, None]
    if isinstance(filename, str):
        with open(join(base_dir, filename)) as fp:
            lines = fp.readlines()
    else:
        lines = filename.readlines()

    meta_lines = [line for line in lines if '-*-' in line]
    args = dict(parse_meta_lines(meta_lines))

    if 'ascii-cols' in args:
        ascii_slice = slice(*args['ascii-cols'])
    else:
        ascii_slice = detect_ascii_slice(lines)

    for chunk in chunker(lines, is_boundary=lambda x: x.strip() == '----'):
        wo_comments = [line for line in chunk if not line.startswith('#')]
        without_ascii = [line[ascii_slice] for line in wo_comments]
        nonempty = [line for line in without_ascii if line.strip()]

        str_bytes = []
        for line in nonempty:
            str_bytes.extend(line.split())

        values = [int(char, 16) for char in str_bytes]

        yield to_bytes(values)
        del str_bytes[:]


def readbytes(filename, base_dir=DATA_DIR):
    # type: (str, str) -> bytes
    packets = readbytes_multiple(filename, base_dir)
    return next(packets)



"""
This module contains the high-level functions to access the library with
asyncio.

Care is taken to make this as pythonic as possible and hide as many of the gory
implementations as possible.

This module provides "syntactic sugar" around the lower-level, but almost
identical, module :py:mod:`puresnmp.aio.api.raw`.

The "raw" module returns the variable types unmodified which are all subclasses
of :py:class:`puresnmp.x690.types.Type`.
"""

# TODO (advanced): This module should not make use of it's own functions. The
#     is beginning to be too "thick", containing too much business logic for a
#     mere abstraction layer.
#     module exists as an abstraction layer only. If one function uses a
#     "siblng" function, valuable information is lost. In general, this module


from __future__ import unicode_literals

import logging
from collections import OrderedDict
from typing import TYPE_CHECKING
from warnings import warn

from . import raw
from ...const import DEFAULT_TIMEOUT
from ...pdu import VarBind
from ...util import BulkResult
from ...x690.types import Type, ObjectIdentifier

if TYPE_CHECKING:  # pragma: no cover
    # pylint: disable=unused-import, invalid-name
    from typing import (
        Any,
        AsyncGenerator,
        Callable,
        Coroutine,
        Dict,
        List,
        Tuple,
        TypeVar,
        Union,
    )
    from puresnmp.typevars import PyType, TWrappedPyType

    TWalkResponse = AsyncGenerator[VarBind, None]
    TFetcher = Callable[[str, str, List[str], int, int],
                        Coroutine[Any, Any, List[VarBind]]]
    T = TypeVar('T', bound=PyType)  # pylint: disable=invalid-name

try:
    unicode = unicode  # type: Callable[[Any], str]
except NameError:
    # pylint: disable=invalid-name
    unicode = str  # type: Callable[[Any], str]

_set = set
LOG = logging.getLogger(__name__)
OID = ObjectIdentifier.from_string

async def get(ip, community, oid, port=161, timeout=DEFAULT_TIMEOUT):
    # type: (str, str, str, int, int) -> PyType
    """
    Delegates to :py:func:`~puresnmp.aio.api.raw.get` but returns simple Python
    types.

    See the "raw" equivalent for detailed documentation & examples.
    """
    raw_value = await raw.get(ip, community, oid, port, timeout=timeout)
    return raw_value.pythonize()  # type: ignore


async def multiget(ip, community, oids, port=161, timeout=DEFAULT_TIMEOUT):
    # type: (str, str, List[str], int, int) -> List[PyType]
    """
    Delegates to :py:func:`~puresnmp.aio.api.raw.multiget` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """
    raw_output = await raw.multiget(ip, community, oids, port, timeout)
    pythonized = [value.pythonize() for value in raw_output]
    return pythonized


async def getnext(ip, community, oid, port=161, timeout=DEFAULT_TIMEOUT):
    # type: (str, str, str, int, int) -> VarBind
    """
    Delegates to :py:func:`~puresnmp.aio.api.raw.getnext` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """
    result = await multigetnext(ip, community, [oid], port, timeout=timeout)
    return result[0]


async def multigetnext(ip, community, oids, port=161, timeout=DEFAULT_TIMEOUT):
    # type: (str, str, List[str], int, int) -> List[VarBind]
    """
    Delegates to :py:func:`~puresnmp.aio.api.raw.multigetnext` but returns
    simple Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """
    raw_output = await raw.multigetnext(ip, community, oids, port, timeout)
    pythonized = [VarBind(oid, value.pythonize())  # type: ignore
                  for oid, value in raw_output]
    return pythonized


async def walk(ip, community, oid, port=161, timeout=DEFAULT_TIMEOUT):
    # type: (str, str, str, int, int) -> TWalkResponse
    """
    Delegates to :py:func:`~puresnmp.aio.api.raw.walk` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """

    raw_result = raw.walk(ip, community, oid, port, timeout)
    async for raw_oid, raw_value in raw_result:
        yield VarBind(raw_oid, raw_value.pythonize())  # type: ignore


async def multiwalk(ip, community, oids, port=161, timeout=DEFAULT_TIMEOUT,
                    fetcher=multigetnext):
    # type: (str, str, List[str], int, int, TFetcher) -> TWalkResponse
    """
    Delegates to :py:func:`~puresnmp.aio.api.raw.multiwalk` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """
    raw_output = raw.multiwalk(ip, community, oids, port, timeout, fetcher)
    async for oid, value in raw_output:
        if isinstance(value, Type):
            value = value.pythonize()
        yield VarBind(oid, value)  # type: ignore


async def set(ip, community, oid, value, port=161, timeout=DEFAULT_TIMEOUT):  # pylint: disable=redefined-builtin
    # type: (str, str, str, Type[T], int, int) -> T
    """
    Delegates to :py:func:`~puresnmp.aio.api.raw.set` but returns simple Python
    types.

    See the "raw" equivalent for detailed documentation & examples.
    """

    result = await multiset(ip, community, [(oid, value)],
                            port, timeout=timeout)
    return result[oid.lstrip('.')]  # type: ignore


async def multiset(ip, community, mappings, port=161, timeout=DEFAULT_TIMEOUT):
    # type: (str, str, List[Tuple[str, Type[TWrappedPyType]]], int, int) -> Dict[str, TWrappedPyType]
    """
    Delegates to :py:func:`~puresnmp.aio.api.raw.multiset` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """

    raw_output = await raw.multiset(ip, community, mappings, port, timeout)
    pythonized = {unicode(oid): value.pythonize()
                  for oid, value in raw_output.items()}
    return pythonized


async def bulkget(ip, community, scalar_oids, repeating_oids, max_list_size=1,
                  port=161, timeout=DEFAULT_TIMEOUT):
    # type: (str, str, List[str], List[str], int, int, int) -> BulkResult
    """
    Delegates to :py:func:`~puresnmp.aio.api.raw.bulkget` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """

    raw_output = await raw.bulkget(ip, community, scalar_oids, repeating_oids,
                                   max_list_size=max_list_size,
                                   port=port,
                                   timeout=timeout)
    pythonized_scalars = {oid: value.pythonize()
                          for oid, value in raw_output.scalars.items()}
    pythonized_list = OrderedDict(
        [(oid, value.pythonize())
         for oid, value in raw_output.listing.items()])
    return BulkResult(pythonized_scalars, pythonized_list)


async def bulkwalk(ip, community, oids, bulk_size=10, port=161,
                   timeout=DEFAULT_TIMEOUT):
    # type: (str, str, List[str], int, int, int) -> TWalkResponse
    """
    Delegates to :py:func:`~puresnmp.aio.api.raw.bulkwalk` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """

    result = multiwalk(
        ip, community, oids, port=port,
        fetcher=raw._bulkwalk_fetcher(  # pylint: disable=protected-access
            bulk_size),
        timeout=timeout)
    async for oid, value in result:
        yield VarBind(oid, value)  # type: ignore


async def table(ip, community, oid, port=161, num_base_nodes=0):
    # type: (str, str, str, int, int) -> AsyncGenerator[Dict[str, Any], None]
    """
    Fetches a table from the SNMP agent. Each value will be converted to a
    pure-python type.

    See :py:func:`puresnmp.api.raw.table` for more information of the returned
    structure.
    """
    if num_base_nodes:
        warn('Usage of "num_base_nodes" in table operations is no longer '
             'required', DeprecationWarning)
    else:
        parsed_oid = OID(oid)
        num_base_nodes = len(parsed_oid) + 1  # type: ignore
    tmp = raw.table(ip, community, oid, port=port,
                    num_base_nodes=num_base_nodes)
    async for row in tmp:
        index = row.pop('0')
        pythonized = {key: value.pythonize() for key, value in row.items()}
        pythonized['0'] = index
        yield pythonized


async def bulktable(ip, community, oid, port=161, num_base_nodes=0, bulk_size=10):
    # type: (str, str, str, int, int, int) -> AsyncGenerator[Dict[str, Any], None]
    """
    Fetch an SNMP table using "bulk" requests converting the values into pure
    Python types.

    See :py:func:`puresnmp.api.raw.table` for more information of the returned
    structure.

    .. versionadded: 1.7.0
    """
    if num_base_nodes:
        warn('Usage of "num_base_nodes" in table operations is no longer '
             'required', DeprecationWarning)
    else:
        parsed_oid = OID(oid)
        num_base_nodes = len(parsed_oid) + 1  # type: ignore
    tmp = raw.bulktable(ip, community, oid, port=port, bulk_size=bulk_size)
    for row in tmp:
        index = row.pop('0')
        pythonized = {key: value.pythonize() for key, value in row.items()}
        pythonized['0'] = index
        yield pythonized



# -*- coding: utf8 -*-
# pylint: skip-file

import six
import sys

try:
    unicode
except NameError:
    unicode = str

from ...x690.util import TypeInfo
from ...x690.types import (
    Boolean,
    Integer,
    Null,
    ObjectIdentifier,
    OctetString,
    Sequence,
    T61String,
    Type,
    UnknownType,
    pop_tlv,
    to_bytes
)

from .. import ByteTester


class TestBoolean(ByteTester):

    def test_encoding_false(self):
        value = Boolean(False)
        result = to_bytes(value)
        expected = b'\x01\x01\x00'
        self.assertBytesEqual(result, expected)

    def test_encoding_true(self):
        value = Boolean(True)
        result = to_bytes(value)
        expected = b'\x01\x01\x01'
        self.assertBytesEqual(result, expected)

    def test_decoding_false(self):
        result = Boolean.from_bytes(b'\x01\x01\x00')
        expected = Boolean(False)
        self.assertEqual(result, expected)

    def test_decoding_true(self):
        result = Boolean.from_bytes(b'\x01\x01\x01')
        expected = Boolean(True)
        self.assertEqual(result, expected)

        result = Boolean.from_bytes(b'\x01\x01\x02')
        expected = Boolean(True)
        self.assertEqual(result, expected)

        result = Boolean.from_bytes(b'\x01\x01\xff')
        expected = Boolean(True)
        self.assertEqual(result, expected)

    def test_pythonize(self):
        result = Boolean(True).pythonize()
        expected = True
        self.assertEqual(result, expected)

    def test_validate_too_long(self):
        """
        Validate what happens when there are too many bytes.
        """
        with six.assertRaisesRegex(self, ValueError, 'Length'):
            Boolean.validate(b'\x01\x00\x00')


class TestObjectIdentifier(ByteTester):

    def setUp(self):
        super(TestObjectIdentifier, self).setUp()
        self.maxDiff = None

    def test_simple_encoding(self):
        """
        A simple OID with no identifier above 127
        """
        oid = ObjectIdentifier(1, 3, 6, 1, 2, 1)
        result = to_bytes(oid)
        expected = b'\x06\x05\x2b\x06\x01\x02\x01'
        self.assertBytesEqual(result, expected)

    def test_simple_decoding(self):
        """
        A simple OID with no identifier above 127
        """
        expected = ObjectIdentifier(1, 3, 6, 1, 2, 1)
        result = ObjectIdentifier.from_bytes(b'\x06\x05\x2b\x06\x01\x02\x01')
        self.assertEqual(result, expected)

    def test_decoding_zero(self):
        """
        A simple OID with the top-level ID '0'
        """
        expected = ObjectIdentifier(0)
        result = ObjectIdentifier.from_bytes(b'\x06\x00')
        self.assertEqual(result, expected)

    def test_encoding_zero(self):
        """
        A simple OID with the top-level ID '0'
        """
        oid = ObjectIdentifier(0)
        result = to_bytes(oid)
        expected = b'\x06\x00'
        self.assertEqual(result, expected)

    def test_multibyte_encoding(self):
        """
        If a sub-identifier has a value bigger than 127, the encoding becomes a
        bit weird. The sub-identifiers are split into multiple sub-identifiers.
        """
        oid = ObjectIdentifier(1, 3, 6, 8072)
        result = to_bytes(oid)
        expected = b'\x06\x04\x2b\x06\xbf\x08'
        self.assertBytesEqual(result, expected)

    def test_multibyte_decoding(self):
        """
        If a sub-identifier has a value bigger than 127, the decoding becomes a
        bit weird. The sub-identifiers are split into multiple sub-identifiers.
        """
        expected = ObjectIdentifier(1, 3, 6, 8072)
        result = ObjectIdentifier.from_bytes(b'\x06\x04\x2b\x06\xbf\x08')
        self.assertEqual(result, expected)

    def test_encode_large_value(self):
        """
        OID sub-identifiers larger than 127 must be split up.

        See https://en.wikipedia.org/wiki/Variable-length_quantity
        """
        result = ObjectIdentifier.encode_large_value(106903)
        expected = [0b10000110, 0b11000011, 0b00010111]
        self.assertEqual(result, expected)

    def test_fromstring(self):
        result = ObjectIdentifier.from_string('1.2.3')
        expected = ObjectIdentifier(1, 2, 3)
        self.assertEqual(result, expected)

    def test_fromstring_leading_dot(self):
        '''
        A leading dot represents the "root" node. This should be allowed as
        string input.
        '''
        result = ObjectIdentifier.from_string('.1.2.3')
        expected = ObjectIdentifier(1, 2, 3)
        self.assertEqual(result, expected)

    def test_pythonize(self):
        result = ObjectIdentifier(1, 2, 3).pythonize()
        expected = '1.2.3'
        self.assertEqual(result, expected)

    def test_str(self):
        result = unicode(ObjectIdentifier(1, 2, 3))
        expected = '1.2.3'
        self.assertEqual(result, expected)

    def test_encode_root(self):
        result = to_bytes(ObjectIdentifier(1))
        expected = b'\x06\x01\x01'
        self.assertBytesEqual(result, expected)

    def test_construct_root_from_string(self):
        """
        Using "." to denote the root OID is common. We should allow this.
        """
        result = ObjectIdentifier.from_string('.')
        expected = ObjectIdentifier(1)
        self.assertEqual(result, expected)

    def test_containment_a(self):
        a = ObjectIdentifier.from_string('1.2.3.4')
        b = ObjectIdentifier.from_string('1.2.3')
        self.assertTrue(a in b)

    def test_containment_b(self):
        a = ObjectIdentifier.from_string('1.2.3.4')
        b = ObjectIdentifier.from_string('1.2.3.4')
        self.assertTrue(a in b)

    def test_containment_c(self):
        a = ObjectIdentifier.from_string('1.3.6.1.2.1.1.1.0')
        b = ObjectIdentifier.from_string('1.3.6.1.2.1')
        self.assertTrue(a in b)

    def test_non_containment_a(self):
        a = ObjectIdentifier.from_string('1.2.3')
        b = ObjectIdentifier.from_string('1.2.3.4')
        self.assertFalse(a in b)

    def test_non_containment_b(self):
        a = ObjectIdentifier.from_string('1.2.3.5')
        b = ObjectIdentifier.from_string('1.2.3.4')
        self.assertFalse(a in b)

    def test_non_containment_c(self):
        a = ObjectIdentifier.from_string('1.2.3.4')
        b = ObjectIdentifier.from_string('1.2.3.5')
        self.assertFalse(a in b)

    def test_non_containment_d(self):
        a = ObjectIdentifier.from_string('1.3.6.1.2.1.25.1.1.0')
        b = ObjectIdentifier.from_string('1.3.6.1.2.1.1.9')
        self.assertFalse(a in b)

    def test_non_containment_e(self):
        a = ObjectIdentifier.from_string('1.3.6.1.2.13')
        b = ObjectIdentifier.from_string('1.3.6.1.2.1')
        self.assertFalse(a in b)

    def test_create_by_iterable(self):
        result = ObjectIdentifier(['1', '2', '3'])
        expected = ObjectIdentifier(1, 2, 3)
        self.assertEqual(result, expected)

    def test_repr(self):
        result = repr(ObjectIdentifier(['1', '2', '3']))
        expected = 'ObjectIdentifier((1, 2, 3))'
        self.assertEqual(result, expected)

    def test_hash(self):
        """
        Test hash function and that it makes sense.
        """
        result = hash(ObjectIdentifier(['1', '2', '3']))
        expected = hash(ObjectIdentifier(1, 2, 3))
        self.assertEqual(result, expected)

    def test_non_containment_f(self):
        """
        This case showed up during development of bulk operations. Throwing it
        into the unit tests to ensure proper containment checks.
        """
        a = ObjectIdentifier(1, 3, 6, 1, 2, 1, 2, 2, 1, 22)
        b = ObjectIdentifier(1, 3, 6, 1, 2, 1, 2, 2, 1, 10, 38)
        self.assertNotIn(a, b, '%s should not be in %s' % (a, b))
        self.assertNotIn(b, a, '%s should not be in %s' % (b, a))

    def test_length_1(self):
        '''
        OIDs with one node should have a length of 1
        '''
        obj = ObjectIdentifier(1)
        self.assertEqual(len(obj), 1)

    def test_length_ge1(self):
        '''
        OIDs with more than one node should have a length equal to the number
        of nodes.
        '''
        obj = ObjectIdentifier(1, 2, 3)
        self.assertEqual(len(obj), 3)

    def test_inequalitites(self):
        a = ObjectIdentifier(1, 2, 3)
        b = ObjectIdentifier(1, 2, 4)
        self.assertTrue(a < b)
        self.assertFalse(b < a)
        self.assertFalse(a < a)
        self.assertFalse(a > b)
        self.assertTrue(b > a)
        self.assertFalse(b > b)

    def test_concatenation(self):
        a = ObjectIdentifier(1, 2, 3)
        b = ObjectIdentifier(4, 5, 6)
        expected = ObjectIdentifier(1, 2, 3, 4, 5, 6)
        result = a + b
        self.assertEqual(result, expected)

    def test_item_access(self):
        a = ObjectIdentifier(1, 2, 3)
        expected = ObjectIdentifier(2)
        result = a[1]
        self.assertEqual(result, expected)


class TestInteger(ByteTester):

    def test_encoding(self):
        value = Integer(100)
        result = to_bytes(value)
        expected = b'\x02\x01\x64'
        self.assertBytesEqual(result, expected)

    def test_decoding(self):
        result = Integer.from_bytes(b'\x02\x01\x0a')
        expected = Integer(10)
        self.assertEqual(result, expected)

    def test_encoding_large_value(self):
        value = Integer(1913359423)
        result = to_bytes(value)
        expected = b"\x02\x04\x72\x0b\x8c\x3f"
        self.assertBytesEqual(result, expected)

    def test_decoding_large_value(self):
        result = Integer.from_bytes(b"\x02\x04\x72\x0b\x8c\x3f")
        expected = Integer(1913359423)
        self.assertEqual(result, expected)

    def test_encoding_zero(self):
        value = Integer(0)
        result = to_bytes(value)
        expected = b"\x02\x01\x00"
        self.assertBytesEqual(result, expected)

    def test_decoding_zero(self):
        result = Integer.from_bytes(b"\x02\x01\x00")
        expected = Integer(0)
        self.assertEqual(result, expected)

    def test_decoding_minus_one(self):
        result = Integer.from_bytes(b"\x02\x01\xff")
        expected = Integer(-1)
        self.assertEqual(result, expected)

    def test_decoding_minus_large_value(self):
        result = Integer.from_bytes(b"\x02\x04\x8d\xf4\x73\xc1")
        expected = Integer(-1913359423)
        self.assertEqual(result, expected)

    def test_pythonize(self):
        result = Integer(1).pythonize()
        expected = 1
        self.assertEqual(result, expected)


class TestIntegerValues(ByteTester):

    def test_32768(self):
        """
        Issue identified in github issue #27

        See https://github.com/exhuma/puresnmp/issues/27
        """
        value = Integer(32768)
        result = to_bytes(value)
        expected = b'\x02\x03\x00\x80\x00'
        self.assertBytesEqual(result, expected)

    def test_minus_one(self):
        value = Integer(-1)
        result = to_bytes(value)
        expected = b'\x02\x01\xff'
        self.assertBytesEqual(result, expected)

    def test_minus_two(self):
        value = Integer(-2)
        result = to_bytes(value)
        expected = b'\x02\x01\xfe'
        self.assertBytesEqual(result, expected)

    def test_zero(self):
        value = Integer(0)
        result = to_bytes(value)
        expected = b'\x02\x01\x00'
        self.assertBytesEqual(result, expected)

    def test_minus_16bit(self):
        value = Integer(-0b1111111111111111)
        result = to_bytes(value)
        expected = b'\x02\x03\xff\x00\x01'
        self.assertBytesEqual(result, expected)

    def test_minus_16bit_plus_one(self):
        value = Integer(-0b1111111111111111 + 1)
        result = to_bytes(value)
        expected = b'\x02\x03\xff\x00\x02'
        self.assertBytesEqual(result, expected)

    def test_minus_16bit_minus_one(self):
        value = Integer(-0b1111111111111111 - 1)
        result = to_bytes(value)
        expected = b'\x02\x03\xff\x00\x00'
        self.assertBytesEqual(result, expected)

    def test_minus_16bit_minus_two(self):
        value = Integer(-0b1111111111111111 - 2)
        result = to_bytes(value)
        expected = b'\x02\x03\xfe\xff\xff'
        self.assertBytesEqual(result, expected)

    def test_16bit(self):
        value = Integer(0b1111111111111111)
        result = to_bytes(value)
        expected = b'\x02\x03\x00\xff\xff'
        self.assertBytesEqual(result, expected)

    def test_16bitplusone(self):
        value = Integer(0b1111111111111111 + 1)
        result = to_bytes(value)
        expected = b'\x02\x03\x01\x00\x00'
        self.assertBytesEqual(result, expected)

    def test_16bitminusone(self):
        value = Integer(0b1111111111111111 - 1)
        result = to_bytes(value)
        expected = b'\x02\x03\x00\xff\xfe'
        self.assertBytesEqual(result, expected)

    def test_32bit(self):
        value = Integer(0b11111111111111111111111111111111)
        result = to_bytes(value)
        expected = b'\x02\x05\x00\xff\xff\xff\xff'
        self.assertBytesEqual(result, expected)


class TestOctetString(ByteTester):

    def test_encoding(self):
        value = OctetString('hello')
        result = to_bytes(value)
        expected = b'\x04\x05hello'
        self.assertBytesEqual(result, expected)

    def test_decoding(self):
        result = OctetString.from_bytes(b'\x04\x05hello')
        expected = OctetString('hello')
        self.assertEqual(result, expected)

    def test_pythonize(self):
        result = OctetString("hello").pythonize()
        expected = b"hello"
        self.assertEqual(result, expected)


class TestT61String(ByteTester):

    def test_encoding(self):
        value = T61String(u'hello Ω')
        result = to_bytes(value)
        expected = b'\x14\x07hello \xe0'
        self.assertBytesEqual(result, expected)

    def test_decoding(self):
        result = T61String.from_bytes(b'\x14\x07hello \xe0')
        expected = T61String(u'hello Ω')
        self.assertEqual(result, expected)

    def test_pythonize_from_string(self):
        obj = T61String(u"hello Ω")
        result = obj.pythonize()
        expected = u"hello Ω"
        self.assertEqual(result, expected)

    def test_pythonize_from_bytes(self):
        obj = T61String(b"hello \xe0")
        result = obj.pythonize()
        expected = u"hello Ω"
        self.assertEqual(result, expected)


class TestSequence(ByteTester):

    def test_encoding(self):
        value = Sequence(
            OctetString('hello'),
            ObjectIdentifier(1, 3, 6),
            Integer(100)
        )
        result = to_bytes(value)
        expected = (
            to_bytes([
                0x30,
                14,  # Expected length (note that an OID drops one byte)
            ]) +
            to_bytes(OctetString('hello')) +
            to_bytes(ObjectIdentifier(1, 3, 6)) +
            to_bytes(Integer(100))
        )
        self.assertBytesEqual(result, expected)

    def test_decoding_simple(self):
        result = Sequence.from_bytes(
            b'\x30\x0b'
            b'\x02\x01\x01'
            b'\x02\x01\x02'
            b'\x04\x03foo'
        )
        expected = Sequence(
            Integer(1),
            Integer(2),
            OctetString('foo'),
        )
        self.assertEqual(result, expected)

    def test_decoding_recursive(self):
        result = Sequence.from_bytes(
            b'\x30\x13'
            b'\x02\x01\x01'
            b'\x02\x01\x02'
            b'\x04\x03foo'
            b'\x30\x06'
            b'\x02\x01\x01'
            b'\x02\x01\x02'
        )
        expected = Sequence(
            Integer(1),
            Integer(2),
            OctetString('foo'),
            Sequence(
                Integer(1),
                Integer(2),
            )
        )
        self.assertEqual(result, expected)

    def test_pythonize(self):
        result = Sequence(Integer(1), Sequence(OctetString('123'))).pythonize()
        expected = [1, [b"123"]]
        self.assertEqual(result, expected)

    def test_iteration(self):
        data = Sequence(
            Integer(1),
            Sequence(OctetString('123')),
            OctetString(b'foo')
        )
        result = [item for item in data]
        expected = [
            Integer(1),
            Sequence(OctetString('123')),
            OctetString(b'foo')
        ]
        self.assertEqual(result, expected)

    def test_indexing(self):
        data = Sequence(
            Integer(1),
            OctetString(b'foo')
        )
        result = data[1]
        expected = OctetString(b'foo')
        self.assertEqual(result, expected)

    def test_repr(self):
        result = repr(Sequence(Integer(10)))
        expected = 'Sequence(Integer(10))'
        self.assertEqual(result, expected)

    def test_length_empty(self):
        result = len(Sequence())
        expected = 0
        self.assertEqual(result, expected)

    def test_length_nonempty(self):
        result = len(Sequence(Integer(1), Integer(2)))
        expected = 2
        self.assertEqual(result, expected)


class TestNull(ByteTester):

    def test_null_is_false(self):
        """
        The Null type should be considered as falsy.
        """
        self.assertFalse(Null())

    def test_validate_true(self):
        Null.validate(b'\x05\x00')

    def test_validate_false(self):
        with self.assertRaises(ValueError):
            Null.validate(b'\x05\x01')

    def test_encoding(self):
        result = to_bytes(Null())
        expected = b'\x05\x00'
        self.assertEqual(result, expected)

    def test_decode_null(self):
        expected = Null()
        result = Null.decode('\x05\x00\x00')
        self.assertEqual(result, expected)

    def test_repr(self):
        expected = 'Null()'
        result = repr(Null())
        self.assertEqual(result, expected)


class TestUnknownType(ByteTester):

    def test_null_from_bytes(self):
        result = UnknownType.from_bytes(b'')
        expected = Null()
        self.assertEqual(result, expected)

    def test_decoding(self):
        result, _ = pop_tlv(b'\x99\x01\x0a')
        expected = UnknownType(0x99, b'\x0a')
        self.assertEqual(result, expected)

    def test_encoding(self):
        result = to_bytes(UnknownType(0x99, b'\x0a'))
        expected = b'\x99\x01\x0a'
        self.assertEqual(result, expected)

    def test_decoding_corrupt_length(self):
        with six.assertRaisesRegex(self, ValueError, 'length'):
            UnknownType.from_bytes(b'\x99\x02\x0a')

    def test_repr(self):
        result = repr(UnknownType(99, b'abc'))
        typeinfo = TypeInfo(u'application', u'constructed', 3)
        if not six.PY2:
            expected = "UnknownType(99, b'abc', typeinfo=%r)" % (typeinfo,)
        else:
            expected = "UnknownType(99, 'abc', typeinfo=%r)" % (typeinfo,)
        self.assertEqual(result, expected)


class TestAllTypes(ByteTester):
    """
    Tests which are valid for all types
    """

    def test_tlv_null(self):
        result = pop_tlv(b'')
        expected = (Null(), b'')
        self.assertEqual(result, expected)

    def test_tlv_simple(self):
        result = pop_tlv(to_bytes([2, 1, 0]))
        expected = (Integer(0), b'')
        self.assertEqual(result, expected)

    def test_tlv_unknown_type(self):
        result = pop_tlv(to_bytes([254, 1, 0]))
        expected = (UnknownType(254, b'\x00'), b'')
        self.assertEqual(result, expected)
        self.assertEqual(result[0].tag, 254)
        self.assertEqual(result[0].length, 1)
        self.assertEqual(result[0].value, b'\x00')

    def test_validation_wrong_typeclass(self):
        with self.assertRaises(ValueError):
            Integer.validate(to_bytes([0b00111110]))

    def test_null_from_bytes(self):
        result = Type.from_bytes(b'')
        expected = Null()
        self.assertEqual(result, expected)

    def test_corrupt_length(self):
        with six.assertRaisesRegex(self, ValueError, 'length'):
            Integer.from_bytes(b'\x02\x01\x01\x01')

    def test_repr(self):
        obj = Type()
        obj.value = 10
        result = repr(obj)
        expected = 'Type(10)'
        self.assertEqual(result, expected)

    def test_childof(self):
        a = ObjectIdentifier(1, 2, 3)
        b = ObjectIdentifier(1, 2, 3, 1)
        c = ObjectIdentifier(1, 2, 4)
        d = ObjectIdentifier(1)
        self.assertTrue(b.childof(a))
        self.assertFalse(a.childof(b))
        self.assertTrue(a.childof(a))
        self.assertFalse(c.childof(a))
        self.assertFalse(a.childof(c))
        self.assertFalse(d.childof(c))
        self.assertTrue(c.childof(d))

    def test_parentdf(self):
        a = ObjectIdentifier(1, 2, 3)
        b = ObjectIdentifier(1, 2, 3, 1)
        c = ObjectIdentifier(1, 2, 4)
        d = ObjectIdentifier(1)
        self.assertFalse(b.parentof(a))
        self.assertTrue(a.parentof(b))
        self.assertTrue(a.parentof(a))
        self.assertFalse(c.parentof(a))
        self.assertFalse(a.parentof(c))
        self.assertTrue(d.parentof(c))
        self.assertFalse(c.parentof(d))



"""
Exceptions for the puresnmp package.
"""
from __future__ import unicode_literals

import socket
from typing import Optional
from puresnmp.const import MAX_VARBINDS

from .x690.types import ObjectIdentifier

# pylint: disable=too-few-public-methods


class SnmpError(Exception):
    """
    Generic exception originating from the puresnmp package. Every SNMP related
    error inherits from this class.
    """


class ErrorResponse(SnmpError):
    """
    A superclass used when the SNMP agent responded with additional error
    information.

    Instances of ``ErrorResponse`` have two attributes concerning the error:

    * ``error_status`` the raw (int) value of the error-status as returned by
      the SNMP agent.
    * ``offending_oid`` the OID identified in the error message which caused
      the error.
    """

    DEFAULT_MESSAGE = 'unknown error'

    @staticmethod
    def construct(error_status, offending_oid, message=''):
        # type: (int, Optional[ObjectIdentifier], str) -> ErrorResponse
        """
        Creates a new instance of an ErrorResponse class, using the proper
        subclass for the given *error_status* value. The message is optional,
        and if not specified, will use the default message for the given class.
        """
        if error_status == 1:
            return TooBig(offending_oid, message)
        if error_status == 2:
            return NoSuchOID(offending_oid, message)
        if error_status == 3:
            return BadValue(offending_oid, message)
        if error_status == 4:
            return ReadOnly(offending_oid, message)
        if error_status == 5:
            return GenErr(offending_oid, message)
        if error_status == 6:
            return NoAccess(offending_oid, message)
        return ErrorResponse(error_status, offending_oid, message)

    def __init__(self, error_status, offending_oid, message=''):
        # type: (int, Optional[ObjectIdentifier], str) -> None
        super(ErrorResponse, self).__init__(
            '%s (status-code: %r) on OID %s' % (
                message or self.DEFAULT_MESSAGE, error_status, offending_oid if offending_oid != None else "unknown"))
        self.error_status = error_status
        self.offending_oid = offending_oid


class TooBig(ErrorResponse):
    """
    This error is returned whenever the size of the generatred response exceeds
    a size-limit defined by the queried device.
    """
    DEFAULT_MESSAGE = 'SNMP response was too big!'
    def __init__(self, offending_oid, message=''):
        # type: (Optional[ObjectIdentifier], str) -> None
        super(TooBig, self).__init__(1, offending_oid)


class NoSuchOID(ErrorResponse):
    """
    This error is returned in any of the following cases:

    * the targeted OID does not support "snmp-set" operations
    * the targeted OID does not exist.
    * the targeted OID is an SMI aggregate type.
    * the targeted OID does not precede a known name in the MIB view.
    """

    DEFAULT_MESSAGE = 'No such name/oid'

    def __init__(self, offending_oid, message=''):
        # type: (Optional[ObjectIdentifier], str) -> None
        super(NoSuchOID, self).__init__(2, offending_oid, message)


class BadValue(ErrorResponse):
    """
    This error is returned whenever a variable is set using an incompatible
    type.
    """

    DEFAULT_MESSAGE = 'Bad value'

    def __init__(self, offending_oid, message=''):
        # type: (Optional[ObjectIdentifier], str) -> None
        super(BadValue, self).__init__(3, offending_oid, message)


class ReadOnly(ErrorResponse):
    """
    This error is returned whenever a variable is set which is not writable.
    """

    DEFAULT_MESSAGE = 'Value is read-only!'

    def __init__(self, offending_oid, message=''):
        # type: (Optional[ObjectIdentifier], str) -> None
        super(ReadOnly, self).__init__(4, offending_oid, message)


class NoAccess(ErrorResponse):
    """
    This error is returned whenever .
    """

    DEFAULT_MESSAGE = 'No Access!'

    def __init__(self, offending_oid, message=''):
        # type: (Optional[ObjectIdentifier], str) -> None
        super(NoAccess, self).__init__(6, offending_oid, message)


class GenErr(ErrorResponse):
    """
    This error is returned for any error which is not covered in the previous
    error classes.
    """

    DEFAULT_MESSAGE = 'General Error (genErr)'

    def __init__(self, offending_oid, message=''):
        # type: (Optional[ObjectIdentifier], str) -> None
        super(GenErr, self).__init__(5, offending_oid, message)


class EmptyMessage(SnmpError):
    """
    Raised when trying to decode an SNMP-Message with no content.
    """


class TooManyVarbinds(SnmpError):
    '''
    Exception which is raised when the number of VarBinds exceeds the limit
    defined in RFC3416.
    device.
    '''

    def __init__(self, num_oids):
        # type: (int) -> None
        super(TooManyVarbinds, self).__init__(
            'Too many VarBinds (%d) in one request!'
            ' RFC3416 limits requests to %d!' % (
                num_oids, MAX_VARBINDS))
        self.num_oids = num_oids


class Timeout(socket.timeout):
    """
    Wrapper for network timeouts.

    This wraps both "socket.timeout" and "asyncio.TimeoutError"
    """

    def __init__(self, message):
        # type: (str) -> None
        super(Timeout, self).__init__(message)
        self.message = message


class FaultySNMPImplementation(SnmpError):
    '''
    Exception which indicates an unexpected response from an SNMP agent.
    '''



'''
Colleciton of utility functions for the puresnmp package.
'''
from typing import TYPE_CHECKING, Any, Dict, List, NamedTuple

if TYPE_CHECKING:
    # pylint: disable=unused-import
    from typing import Optional, Tuple
    from .pdu import VarBind
    from .x690.types import ObjectIdentifier


WalkRow = NamedTuple('WalkRow', [
    ('value', Any),
    ('unfinished', bool),
])
BulkResult = NamedTuple('BulkResult', [
    ('scalars', Dict[str, Any]),
    ('listing', Dict[str, Any]),
])


def group_varbinds(varbinds, effective_roots, user_roots=None):
    # type: (List[VarBind], List[ObjectIdentifier], Optional[List[ObjectIdentifier]]) -> Dict[ObjectIdentifier, List[VarBind]]
    """
    Takes a list of varbinds and a list of base OIDs and returns a mapping from
    those base IDs to lists of varbinds.

    Varbinds returned from a walk operation which targets multiple OIDs is
    returned as "interleaved" list. This functions extracts these interleaved
    items into a more usable dictionary.

    :param varbinds: A list of VarBind instances.
    :param effective_roots: The list of OIDs that were requested from the SNMP
        agent.
    :param user_roots: The set of VarBind instances that were requested by the
        user. This is used internally for walk requests. On each request
        following the first, the requested OIDs will differ from the OIDs
        requested by the user. This list will keep track of the original OIDs
        to determine when the walk needs to terminate.
    """
    user_roots = user_roots or []
    n = len(effective_roots)

    results = {}
    for i in range(n):
        results[effective_roots[i]] = varbinds[i::n]

    if user_roots:
        new_results = {}
        for key, value in results.items():
            containment = [base for base in user_roots if key in base]
            if len(containment) > 1:
                raise RuntimeError('Unexpected OID result. A value was '
                                   'contained in more than one base than '
                                   'should be possible!')
            if not containment:
                continue
            new_results[containment[0]] = value
            results = new_results

    return results


def get_unfinished_walk_oids(grouped_oids):
    # type: (Dict[ObjectIdentifier, List[VarBind]]) -> List[Tuple[ObjectIdentifier, WalkRow]]
    """
    :param grouped_oids: A dictionary containing VarBinds as values. The keys
        are the base OID of those VarBinds as requested by the user. We need to
        keep track of the base to be able to tell when a walk over OIDs is
        finished (that is, when we hit the first OID outside the base).
    """

    # grouped_oids contains a list of values for each requested OID. We need to
    # determine if we need to continue fetching: Inspect the last item of each
    # list if those OIDs are still children of the requested IDs we need to
    # continue fetching using *those* IDs (as we're using GetNext behaviour).
    # If they are *not* children of the requested OIDs, we went too far (in the
    # case of a bulk operation) and need to remove all outliers.
    #
    # The above behaviour is the same for both bulk and simple operations. For
    # simple operations we simply have a list of 1 element per OID, but the
    # behaviour is identical

    # Build a mapping from the originally requested OID to the last fetched OID
    # from that tree.
    last_received_oids = {k: WalkRow(v[-1], v[-1].oid in k)  # type: ignore
                          for k, v in grouped_oids.items() if v}

    output = [item for item in sorted(last_received_oids.items())
              if item[1].unfinished]
    return output



"""
Low-Level network transport.

This module mainly exist to enable a "seam" for mocking/patching out during
testing.

The module is excluded from coverage. It contains all the "dirty" stuff that's
hard to test.
"""

# TODO (beginner, no-dev): Ignore this file from coverage without adding
#                          "pragma: no cover" to each function.

import logging
import socket
from ipaddress import ip_address
from typing import TYPE_CHECKING, Generator

from .exc import Timeout
from .typevars import SocketInfo, SocketResponse
from .x690.util import visible_octets

LOG = logging.getLogger(__name__)

#: The default number of retries for UDP packets
RETRIES = 3

#: Low-level socket buffer-size. If you run into timeouts you may want to
#: increase this
BUFFER_SIZE = 4096  # 4 KiB


if TYPE_CHECKING:
    from typing import Optional


class Transport(object):
    """
    A simple UDP transport.

    Calling ``send`` will attempt to send a packet as many times as specified
    in *retries* (default=3). If it fails after those attemps, a
    py:exc:`puresnmp.Timeout` exception is raised.

    :param timeout: How long to wait on the socket before retrying
    :param retries: The number of retries attempted if a low-level
        socket-timeout occurs. If set to ``None`` it will use
        :py:data:`puresnmp.transport.RETRIES` as default.
    :param buffer_size: How much data to read from the socket. If this is too
        small, it may result in incomplete (corrupt) packages. This should be
        kept as low as possible (see ``man(2) recv``). If set to ``None`` it
        will use :py:data:`puresnmp.transport.BUFFER_SIZE` as default.
    """

    def __init__(self, timeout=2, retries=None, buffer_size=None):
        # type: (int, Optional[int], Optional[int]) -> None
        self.timeout = timeout
        self.retries = retries or RETRIES
        self.buffer_size = buffer_size or BUFFER_SIZE

    def send(
            self, ip, port, packet):  # pragma: no cover
        # type: ( str, int, bytes ) -> bytes
        """
        Opens a TCP connection to *ip:port*, sends a packet with *bytes* and
        returns the raw bytes as returned from the remote host.
        """
        checked_ip = ip_address(ip)
        if checked_ip.version == 4:
            address_family = socket.AF_INET
        else:
            address_family = socket.AF_INET6

        sock = socket.socket(address_family, socket.SOCK_DGRAM)
        sock.settimeout(self.timeout)

        for num_retry in range(self.retries):
            try:
                if LOG.isEnabledFor(logging.DEBUG):
                    hexdump = visible_octets(packet)
                    LOG.debug('Sending packet to %s:%s (attempt %d/%d)\n%s',
                              ip, port, (num_retry+1), self.retries, hexdump)
                sock.sendto(packet, (ip, port))
                response = sock.recv(self.buffer_size)
                break
            except socket.timeout:
                LOG.debug('Timeout during attempt #%d',
                          (num_retry+1))  # TODO add detail
                continue
        else:
            sock.close()
            raise Timeout(
                "Max of %d retries reached" % self.retries)  # type: ignore
        sock.close()

        if LOG.isEnabledFor(logging.DEBUG):
            hexdump = visible_octets(response)
            LOG.debug('Received packet:\n%s', hexdump)

        return response

    def listen(self, bind_address='0.0.0.0', port=162):  # pragma: no cover
        # type: (str, int) -> Generator[SocketResponse, None, None]
        """
        Sets up a listening UDP socket and returns a generator over recevied
        packets::

            >>> transport = Transport()
            >>> for seq, packet in enumerate(transport.listen()):
            ...     print(seq, repr(packet))
            0, b'...'
            1, b'...'
            2, b'...'

        .. note::

            This defaults to the standard SNMP Trap port 162. This is a
            privileged port so processes using this port must run as root!
        """
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            sock.bind((bind_address, port))
            while True:
                request, addr = sock.recvfrom(self.buffer_size)
                if LOG.isEnabledFor(logging.DEBUG):
                    hexdump = visible_octets(request)
                    LOG.debug('Received packet:\n%s', hexdump)

                yield SocketResponse(request, SocketInfo(addr[0], addr[1]))

    def get_request_id(self):  # pragma: no cover
        # type: () -> int
        """
        Generates a SNMP request ID. This value should be unique for each
        request.
        """
        # TODO check if this is good enough. My gut tells me "no"! Depends if
        # it has to be unique across all clients, or just one client. If it's
        # just one client it *may* be enough.
        from time import time
        return int(time())



# pylint: skip-file

import sys

import six

from .. import ByteTester
from ...x690.types import ObjectIdentifier
from ...x690.util import (
    Length,
    TypeInfo,
    decode_length,
    encode_length,
    tablify,
    to_bytes,
    visible_octets
)

if sys.version_info > (3, 3):
    from unittest.mock import patch
else:
    from mock import patch



class TestTypeInfoDecoding(ByteTester):
    """
    Tests the various possible combinations for decoding type-hint octets into
    Python objects.
    """

    def test_from_bytes_raw_value(self):
        result = TypeInfo.from_bytes(0b00011110)._raw_value
        expected = 0b00011110
        self.assertEqual(result, expected)

    def test_from_bytes_a(self):
        result = TypeInfo.from_bytes(0b00011110)
        expected = TypeInfo(TypeInfo.UNIVERSAL, TypeInfo.PRIMITIVE, 0b11110)
        self.assertEqual(result, expected)

    def test_from_bytes_b(self):
        result = TypeInfo.from_bytes(0b00111110)
        expected = TypeInfo(TypeInfo.UNIVERSAL, TypeInfo.CONSTRUCTED, 0b11110)
        self.assertEqual(result, expected)

    def test_from_bytes_c(self):
        result = TypeInfo.from_bytes(0b01011110)
        expected = TypeInfo(TypeInfo.APPLICATION, TypeInfo.PRIMITIVE, 0b11110)
        self.assertEqual(result, expected)

    def test_from_bytes_d(self):
        result = TypeInfo.from_bytes(0b01111110)
        expected = TypeInfo(TypeInfo.APPLICATION, TypeInfo.CONSTRUCTED, 0b11110)
        self.assertEqual(result, expected)

    def test_from_bytes_e(self):
        result = TypeInfo.from_bytes(0b10011110)
        expected = TypeInfo(TypeInfo.CONTEXT, TypeInfo.PRIMITIVE, 0b11110)
        self.assertEqual(result, expected)

    def test_from_bytes_f(self):
        result = TypeInfo.from_bytes(0b10111110)
        expected = TypeInfo(TypeInfo.CONTEXT, TypeInfo.CONSTRUCTED, 0b11110)
        self.assertEqual(result, expected)

    def test_from_bytes_g(self):
        result = TypeInfo.from_bytes(0b11011110)
        expected = TypeInfo(TypeInfo.PRIVATE, TypeInfo.PRIMITIVE, 0b11110)
        self.assertEqual(result, expected)

    def test_from_bytes_h(self):
        result = TypeInfo.from_bytes(0b11111110)
        expected = TypeInfo(TypeInfo.PRIVATE, TypeInfo.CONSTRUCTED, 0b11110)
        self.assertEqual(result, expected)


class TestTypeInfoEncoding(ByteTester):
    """
    Tests the various possible combinations for encoding type-hint instances
    into bytes.
    """

    def test_to_bytes_a(self):
        obj = TypeInfo(TypeInfo.UNIVERSAL, TypeInfo.PRIMITIVE, 0b11110)
        result = to_bytes(obj)
        expected = to_bytes([0b00011110])
        self.assertEqual(result, expected)

    def test_to_bytes_b(self):
        obj = TypeInfo(TypeInfo.UNIVERSAL, TypeInfo.CONSTRUCTED, 0b11110)
        result = to_bytes(obj)
        expected = to_bytes([0b00111110])
        self.assertEqual(result, expected)

    def test_to_bytes_c(self):
        obj = TypeInfo(TypeInfo.APPLICATION, TypeInfo.PRIMITIVE, 0b11110)
        result = to_bytes(obj)
        expected = to_bytes([0b01011110])
        self.assertEqual(result, expected)

    def test_to_bytes_d(self):
        obj = TypeInfo(TypeInfo.APPLICATION, TypeInfo.CONSTRUCTED, 0b11110)
        result = to_bytes(obj)
        expected = to_bytes([0b01111110])
        self.assertEqual(result, expected)

    def test_to_bytes_e(self):
        obj = TypeInfo(TypeInfo.CONTEXT, TypeInfo.PRIMITIVE, 0b11110)
        result = to_bytes(obj)
        expected = to_bytes([0b10011110])
        self.assertEqual(result, expected)

    def test_to_bytes_f(self):
        obj = TypeInfo(TypeInfo.CONTEXT, TypeInfo.CONSTRUCTED, 0b11110)
        result = to_bytes(obj)
        expected = to_bytes([0b10111110])
        self.assertEqual(result, expected)

    def test_to_bytes_g(self):
        obj = TypeInfo(TypeInfo.PRIVATE, TypeInfo.PRIMITIVE, 0b11110)
        result = to_bytes(obj)
        expected = to_bytes([0b11011110])
        self.assertEqual(result, expected)

    def test_to_bytes_h(self):
        obj = TypeInfo(TypeInfo.PRIVATE, TypeInfo.CONSTRUCTED, 0b11110)
        result = to_bytes(obj)
        expected = to_bytes([0b11111110])
        self.assertEqual(result, expected)

    def test_to_bytes_error(self):
        obj = TypeInfo(TypeInfo.PRIVATE, TypeInfo.CONSTRUCTED, 0b11110)
        with patch('puresnmp.x690.util.bytes') as ptch:
            ptch.side_effect = TypeError('booh')
            with six.assertRaisesRegex(self, TypeError, r'booh'):
                to_bytes(obj)


class TestTypeInfoClass(ByteTester):
    """
    Tests various "implied" functionality of TypeInfo objects.
    """

    def test_equality(self):
        a = TypeInfo(TypeInfo.UNIVERSAL, TypeInfo.CONSTRUCTED, 0b11110)
        b = TypeInfo(TypeInfo.UNIVERSAL, TypeInfo.CONSTRUCTED, 0b11110)
        self.assertEqual(a, b)

    def test_inequality(self):
        a = TypeInfo(TypeInfo.UNIVERSAL, TypeInfo.CONSTRUCTED, 0b11110)
        b = TypeInfo(TypeInfo.UNIVERSAL, TypeInfo.CONSTRUCTED, 0b10110)
        self.assertNotEqual(a, b)

    def test_encoding_symmetry_a(self):
        """
        Encoding an object to bytes, and then decoding the resulting bytes
        should yield the same instance.
        """
        expected = TypeInfo(TypeInfo.UNIVERSAL, TypeInfo.CONSTRUCTED, 0b11110)
        result = TypeInfo.from_bytes(to_bytes(expected))
        self.assertEqual(result, expected)

    def test_dencoding_symmetry_b(self):
        """
        Decoding an object from bytes, and then encoding the resulting instance
        should yield the same bytes.
        """
        expected = to_bytes([0b11111110])
        result = to_bytes(TypeInfo.from_bytes(expected))
        self.assertEqual(result, expected)

    def test_impossible_class(self):
        instance = TypeInfo(10, 100, 1000)
        with six.assertRaisesRegex(self, ValueError, 'class'):
            to_bytes(instance)

    def test_impossible_pc(self):
        instance = TypeInfo(TypeInfo.APPLICATION, 100, 1000)
        with six.assertRaisesRegex(self, ValueError, 'primitive/constructed'):
            to_bytes(instance)


class TestLengthOctets(ByteTester):

    def test_encode_length_short(self):
        expected = to_bytes([0b00100110])
        result = encode_length(38)
        self.assertEqual(result, expected)

    def test_encode_length_long(self):
        expected = to_bytes([0b10000001, 0b11001001])
        result = encode_length(201)
        self.assertBytesEqual(result, expected)

    def test_encode_length_longer(self):
        expected = to_bytes([0b10000010, 0b00000001, 0b00101110])
        result = encode_length(302)
        self.assertBytesEqual(result, expected)

    def test_encode_length_longer_2(self):
        expected = to_bytes([0x81, 0xa4])
        result = encode_length(164)
        self.assertBytesEqual(result, expected)

    def test_encode_length_indefinite(self):
        expected = to_bytes([0b10000000])
        result = encode_length(Length.INDEFINITE)
        self.assertBytesEqual(result, expected)

    def test_identifier_long(self):
        with self.assertRaises(NotImplementedError):
            TypeInfo.from_bytes(0b11111111)
        self.skipTest('Not yet implemented')  # TODO implement

    def test_decode_length_short(self):
        data = b'\x05'
        expected = 5
        result, data = decode_length(data)
        self.assertEqual(result, expected)
        self.assertEqual(data, b'')

    def test_decode_length_long(self):
        data = to_bytes([0b10000010, 0b00000001, 0b10110011])
        expected = 435
        result, data = decode_length(data)
        self.assertEqual(result, expected)
        self.assertEqual(data, b'')

    def test_decode_length_longer(self):
        data = to_bytes([0x81, 0xa4])
        expected = 164
        result, data = decode_length(data)
        self.assertEqual(result, expected)
        self.assertEqual(data, b'')

    def test_decode_length_indefinite(self):
        with self.assertRaises(NotImplementedError):
            decode_length(to_bytes([0b10000000]))

    def test_decode_length_reserved(self):
        with self.assertRaises(NotImplementedError):
            decode_length(to_bytes([0b11111111]))


class TestHelpers(ByteTester):

    def test_visible_octets_minimal(self):
        result = visible_octets(to_bytes([0b00000000, 0b01010101]))
        expected = '00 55                                              .U'
        self.assertEqual(result, expected)

    def test_visible_octets_double_space(self):
        """
        Test that we have a double space after 8 octets for better readability
        """
        result = visible_octets(to_bytes([
            0b00000000,
            0b01010101,
            0b00000000,
            0b01010101,
            0b00000000,
            0b01010101,
            0b00000000,
            0b01010101,
            0b01010101,
        ]))
        expected = ('00 55 00 55 00 55 00 55  55                        '
                    '.U.U.U.UU')
        self.assertEqual(result, expected)

    def test_visible_octets_multiline(self):
        """
        If we have more than 16 octets, we need to go to a new line.
        """
        result = visible_octets(to_bytes([0b00000000, 0b01010101] * 9))
        expected = ('00 55 00 55 00 55 00 55  00 55 00 55 00 55 00 55   '
                    '.U.U.U.U.U.U.U.U\n'
                    '00 55                                              '
                    '.U')
        self.assertEqual(result, expected)

    def test_tablify_simple(self):
        data = [
            (ObjectIdentifier.from_string('1.2.1.1'), 'row 1 col 1'),
            (ObjectIdentifier.from_string('1.2.1.2'), 'row 2 col 1'),
            (ObjectIdentifier.from_string('1.2.2.1'), 'row 1 col 2'),
            (ObjectIdentifier.from_string('1.2.2.2'), 'row 2 col 2'),
        ]
        result = tablify(data)
        expected = [
            {'0': '1',
             '1': 'row 1 col 1',
             '2': 'row 1 col 2'},
            {'0': '2',
             '1': 'row 2 col 1',
             '2': 'row 2 col 2'},
        ]
        six.assertCountEqual(self, result, expected)

    def test_tablify_with_base(self):
        """
        Sometimes, the row indices are actually OIDs, so we need a way to "cut"
        these off.
        """
        data = [
            (ObjectIdentifier.from_string('1.2.1.1.1.1'), 'row 1.1.1 col 1'),
            (ObjectIdentifier.from_string('1.2.1.2.1.1'), 'row 2.1.1 col 1'),
            (ObjectIdentifier.from_string('1.2.2.1.1.1'), 'row 1.1.1 col 2'),
            (ObjectIdentifier.from_string('1.2.2.2.1.1'), 'row 2.1.1 col 2'),
        ]
        result = tablify(data, num_base_nodes=2)
        expected = [
            {'0': '1.1.1',
             '1': 'row 1.1.1 col 1',
             '2': 'row 1.1.1 col 2'},
            {'0': '2.1.1',
             '1': 'row 2.1.1 col 1',
             '2': 'row 2.1.1 col 2'},
        ]
        six.assertCountEqual(self, result, expected)

    def test_tablify_with_base_oid(self):
        """
        Using "tablify" with "num_base_nodes" is not very intuitive. It is
        easier to pass in the base-oid instead.
        """
        OID = ObjectIdentifier.from_string
        table_entry_oid = '1.3.6.1.2.1.1.9.1'

        # Note: This is not a real table. For the test, the row-id suffix ".1"
        # was added
        data = [
            (OID('1.3.6.1.2.1.1.9.1.2.1.1'), 'row 1.1 col 2'),
            (OID('1.3.6.1.2.1.1.9.1.2.2.1'), 'row 2.1 col 2'),
            (OID('1.3.6.1.2.1.1.9.1.3.1.1'), 'row 1.1 col 3'),
            (OID('1.3.6.1.2.1.1.9.1.3.2.1'), 'row 2.1 col 3'),
            (OID('1.3.6.1.2.1.1.9.1.4.1.1'), 'row 1.1 col 4'),
            (OID('1.3.6.1.2.1.1.9.1.4.2.1'), 'row 2.1 col 4'),
        ]

        result = tablify(data, base_oid=table_entry_oid)
        expected = [
            {
                '0': '1.1',
                '2': 'row 1.1 col 2',
                '3': 'row 1.1 col 3',
                '4': 'row 1.1 col 4',
            }, {
                '0': '2.1',
                '2': 'row 2.1 col 2',
                '3': 'row 2.1 col 3',
                '4': 'row 2.1 col 4',
            },
        ]
        six.assertCountEqual(self, result, expected)

    def test_tmp(self):
        data = [
            (ObjectIdentifier.from_string('1.2.1.5.10'), 'row 5.10 col 1'),
            (ObjectIdentifier.from_string('1.2.1.6.10'), 'row 6.10 col 1'),
            (ObjectIdentifier.from_string('1.2.2.5.10'), 'row 5.10 col 2'),
            (ObjectIdentifier.from_string('1.2.2.6.10'), 'row 6.10 col 2'),
        ]
        result = tablify(data, num_base_nodes=2)
        expected = [
            {'0': '5.10', '1': 'row 5.10 col 1', '2': 'row 5.10 col 2'},
            {'0': '6.10', '1': 'row 6.10 col 1', '2': 'row 6.10 col 2'},
        ]
        six.assertCountEqual(self, result, expected)


class TestGithubIssue23(ByteTester):
    """
    In issue #23 a problem was raised that the byte-order was incorrect when
    encoding large values for "length" information.

    This test-case takes the value (435) from the X.690 Wikipedia page as
    reference and has indeed highlighted the error.
    """

    def test_encode(self):
        expected = to_bytes([0b10000010, 0b00000001, 0b10110011])
        result = encode_length(435)
        self.assertBytesEqual(result, expected)

    def test_decode(self):
        data = to_bytes([0b10000010, 0b00000001, 0b10110011])
        expected = 435
        result, data = decode_length(data)
        self.assertEqual(result, expected)
        self.assertEqual(data, b'')

    def test_symmetry(self):
        result, _ = decode_length(encode_length(435))
        self.assertEqual(result, 435)



# pylint: skip-file
from __future__ import unicode_literals

from datetime import timedelta
from ipaddress import ip_address

from . import ByteTester
from .. import types as t
from ..x690.util import to_bytes


class TestIpAddress(ByteTester):

    def test_decoding(self):
        result = t.IpAddress.from_bytes(b'\x40\x04\x80\x96\xa1\x09')
        expected = t.IpAddress(b'\x80\x96\xa1\x09')
        self.assertEqual(result, expected)

    def test_encoding(self):
        value = t.IpAddress(b'\x80\x96\xa1\x09')
        result = to_bytes(value)
        expected = b'\x40\x04\x80\x96\xa1\x09'
        self.assertBytesEqual(result, expected)

    def test_conversion_to_python(self):
        result = t.IpAddress(b'\x80\x96\xa1\x09').pythonize()
        expected = b'\x80\x96\xa1\x09'
        # TODO 2.0.0 expected = ip_address('128.150.161.9')
        self.assertEqual(result, expected)

    def test_conversion_from_python(self):
        input = ip_address('128.150.161.9')
        result = t.IpAddress(input)
        expected = t.IpAddress(b'\x80\x96\xa1\x09')
        self.assertEqual(result, expected)

    def test_conversion_symmetry(self):
        input = b'\x80\x96\xa1\t'
        # TODO 2.0.0 input = ip_address('128.150.161.9')
        result = t.IpAddress(input).pythonize()
        self.assertEqual(result, input)


class TestTimeTicks(ByteTester):

    def test_decoding(self):
        result = t.TimeTicks.from_bytes(b'\x43\x01\x0a')
        expected = t.TimeTicks(10)
        self.assertEqual(result, expected)

    def test_encoding(self):
        value = t.TimeTicks(100)
        result = to_bytes(value)
        expected = b'\x43\x01\x64'
        self.assertBytesEqual(result, expected)

    def test_conversion_to_python(self):
        # NOTE: 41 centiseconds = 410000 microseconds
        result = t.TimeTicks(18890041).pythonize()
        expected = timedelta(days=2, hours=4, minutes=28, seconds=20,
                             microseconds=410000)
        self.assertEqual(result, expected)

    def test_conversion_from_python(self):
        # NOTE: 41 centiseconds = 410000 microseconds
        input = timedelta(days=2, hours=4, minutes=28, seconds=20,
                          microseconds=410000)
        result = t.TimeTicks(input)
        expected = t.TimeTicks(18890041)
        self.assertEqual(result, expected)

    def test_conversion_symmetry(self):
        input = timedelta(days=2, hours=4, minutes=28, seconds=20,
                          microseconds=410000)
        result = t.TimeTicks(input).pythonize()
        self.assertEqual(result, input)



from puresnmp.test import readbytes
from puresnmp.pdu import PDU
from puresnmp.x690.types import OctetString, Integer

from puresnmp.api.raw import bulkget
import pytest

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch  # type: ignore


def test_54_endofmibview():
    data = readbytes('gh-issues/54-endofmibview.hex')
    with patch('puresnmp.api.raw.Transport') as ptch:
        ptch().send.return_value = data
        ptch().get_request_id.return_value = 123
        result = bulkget('192.0.2.1', 'private', [], ['1.2.3'],
                 max_list_size=10)
    assert result.scalars == {}

    expected_lists = {
        '1.3.6.1.4.1.12325.1.303.1.9.1.3.1.0': OctetString(b'\x00'*404),
        '1.3.6.1.4.1.12325.1.303.1.9.1.3.2.0': OctetString(b'\x00'*404),
        '1.3.6.1.6.3.1.1.6.1.0': Integer(1103779487),
        '1.3.6.1.6.3.10.2.1.1.0': OctetString(b'\x80\x00\x08\xbe\x803740773355'),
        '1.3.6.1.6.3.10.2.1.2.0': Integer(1),
        '1.3.6.1.6.3.10.2.1.3.0': Integer(632),
        '1.3.6.1.6.3.10.2.1.4.0': Integer(1500),
    }
    assert result.listing == expected_lists
